
__docformat__ = "restructuredtext en"
__all__ = ['CropWidget']

from typing import Optional, Sequence

import numpy as np
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QMainWindow, QVBoxLayout, QPushButton, QLabel, QFrame
from vtkmodules import vtkCommonMath
from vtkmodules.vtkInteractionWidgets import vtkBoxWidget2, vtkBoxRepresentation

from opentps.core.data.images._image3D import Image3D
from opentps.core.processing.imageProcessing.resampler3D import crop3DDataAroundBox
from opentps.gui.panels.patientDataPanel.patientDataSelection import PatientDataSelection
from opentps.gui.viewer.dataViewerComponents.image3DViewer import Image3DViewer
from opentps.gui.viewer.dataViewer import DroppedObject


class CropWidget(QMainWindow):
    def __init__(self, viewController, parent=None):
        super().__init__(parent)

        self._viewController = viewController

        self.setWindowTitle('Crop tool')
        self.resize(800, 600)

        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        self._mainLayout = QHBoxLayout()
        centralWidget.setLayout(self._mainLayout)

        self._viewers = ThreeViewsGrid(self._viewController)

        self._dragLabel = QLabel("Select as many images as you want to crop.\nDrop the image needed to create a box.", self)
        self._dragLabel.setWordWrap(True)

        self._cropDataButton = QPushButton('Crop all selected data')
        self._cropDataButton.clicked.connect(self._cropData)

        self._dataSelection = PatientDataSelection(self._viewController, self)

        self._menuFrame = QFrame(self)
        self._menuFrame.setFixedWidth(200)
        self._menuLayout = QVBoxLayout(self._menuFrame)
        self._menuFrame.setLayout(self._menuLayout)

        self._mainLayout.addWidget(self._menuFrame)
        self._menuLayout.addWidget(self._dragLabel)
        self._menuLayout.addWidget(self._dataSelection)
        self._menuLayout.addWidget(self._cropDataButton)
        self._mainLayout.addWidget(self._viewers)

    def closeEvent(self, event):
        self._viewers.close()
        super().closeEvent(event)

    def _cropData(self):
        selectedData = self._dataSelection.selectedData

        box = self._viewers.getBoundingBox()

        xx = np.array([box[0], box[1]])
        yy = np.array([box[2], box[3]])
        zz = np.array([box[4], box[5]])

        box = [[xx.min(), xx.max()], [yy.min(), yy.max()], [zz.min(), zz.max()]]

        for data in selectedData:
            crop3DDataAroundBox(data, box, marginInMM=[0, 0, 0])


class ThreeViewsGrid(QWidget):
    def __init__(self, viewController):
        super().__init__()

        self._viewController = viewController

        self._mainLayout = QHBoxLayout(self)
        self.setLayout(self._mainLayout)

        self._viewer0 = Image3DViewer(viewController)
        self._viewer1 = Image3DViewer(viewController)
        self._viewer2 = Image3DViewer(viewController)

        self._viewer0.viewType = Image3DViewer.ViewerTypes.CORONAL
        self._viewer1.viewType = Image3DViewer.ViewerTypes.AXIAL
        self._viewer2.viewType = Image3DViewer.ViewerTypes.SAGITTAL

        self._viewer0.crossHairEnabled = True
        self._viewer1.crossHairEnabled = True
        self._viewer2.crossHairEnabled = True

        self._mainLayout.addWidget(self._viewer0)
        self._mainLayout.addWidget(self._viewer1)
        self._mainLayout.addWidget(self._viewer2)

        self.setAcceptDrops(True)
        self.dragEnterEvent = lambda event: event.accept()
        self.dropEvent = lambda event: self._dropEvent(event)

    def close(self):
        self._viewer0.close()
        self._viewer1.close()
        self._viewer2.close()

    def _dropEvent(self, e):
        if e.mimeData().hasText():
            droppedIsImage = e.mimeData().text() == DroppedObject.DropTypes.IMAGE

            if droppedIsImage:
                e.accept()
                self._setMainImage(self._viewController.selectedImage)
                return
        e.ignore()


    def _setMainImage(self, image: Optional[Image3D]):
        if isinstance(image, Image3D) or (image is None):
            self._viewer0.primaryImage = image
            self._viewer1.primaryImage = image
            self._viewer2.primaryImage = image

            interactor0 = self._viewer0._renderWindow.GetInteractor()
            interactor1 = self._viewer1._renderWindow.GetInteractor()
            interactor2 = self._viewer2._renderWindow.GetInteractor()

            self.boxRep0 = vtkBoxRepresentation()
            self.boxRep0.SetPlaceFactor(1)
            self.boxRep0.PlaceWidget(self._viewer0._primaryImageLayer._mainActor.GetBounds())

            self.boxRep1 = vtkBoxRepresentation()
            self.boxRep1.SetPlaceFactor(1)
            self.boxRep1.PlaceWidget(self._viewer1._primaryImageLayer._mainActor.GetBounds())

            self.boxRep2 = vtkBoxRepresentation()
            self.boxRep2.SetPlaceFactor(1)
            self.boxRep2.PlaceWidget(self._viewer2._primaryImageLayer._mainActor.GetBounds())


            self.boxWidget0 = vtkBoxWidget2()
            self.boxWidget0.SetRotationEnabled(False)
            self.boxWidget0.SetInteractor(interactor0)
            self.boxWidget0.SetRepresentation(self.boxRep0)
            self.boxWidget0.On()

            self.boxWidget1 = vtkBoxWidget2()
            self.boxWidget1.SetRotationEnabled(False)
            self.boxWidget1.SetInteractor(interactor1)
            self.boxWidget1.SetRepresentation(self.boxRep1)
            self.boxWidget1.On()

            self.boxWidget2 = vtkBoxWidget2()
            self.boxWidget2.SetRotationEnabled(False)
            self.boxWidget2.SetInteractor(interactor2)
            self.boxWidget2.SetRepresentation(self.boxRep2)
            self.boxWidget2.On()

            self._viewer0._renderWindow.Render()
            self._viewer1._renderWindow.Render()
            self._viewer2._renderWindow.Render()

            interactor0.Start()
            interactor1.Start()
            interactor2.Start()

            self.boxWidget0.AddObserver('InteractionEvent', self.onInteraction0)
            self.boxWidget1.AddObserver('InteractionEvent', self.onInteraction1)
            self.boxWidget2.AddObserver('InteractionEvent', self.onInteraction2)

    def onInteraction0(self, obj, event):
        self.boxRep1.PlaceWidget(self._2DboundsFromViewerToViewer(self._viewer0, self.boxRep0, self._viewer1, self.boxRep1))
        self.boxRep2.PlaceWidget(self._2DboundsFromViewerToViewer(self._viewer0, self.boxRep0, self._viewer2, self.boxRep2))
        self._viewer1._renderWindow.Render()
        self._viewer2._renderWindow.Render()

    def onInteraction1(self, obj, event):
        self.boxRep0.PlaceWidget(self._2DboundsFromViewerToViewer(self._viewer1, self.boxRep1, self._viewer0, self.boxRep0))
        self.boxRep2.PlaceWidget(self._2DboundsFromViewerToViewer(self._viewer1, self.boxRep1, self._viewer2, self.boxRep2))
        self._viewer0._renderWindow.Render()
        self._viewer2._renderWindow.Render()

    def onInteraction2(self, obj, event):
        self.boxRep0.PlaceWidget(self._2DboundsFromViewerToViewer(self._viewer2, self.boxRep2, self._viewer0, self.boxRep0))
        self.boxRep1.PlaceWidget(self._2DboundsFromViewerToViewer(self._viewer2, self.boxRep2, self._viewer1, self.boxRep1))
        self._viewer0._renderWindow.Render()
        self._viewer1._renderWindow.Render()

    def _2DboundsFromViewerToViewer(self, viewer0:Image3DViewer, boxRep0, viewer1:Image3DViewer, boxRep1) -> Sequence[float]:
        bounds = self._boundsFromViewerToViewer(viewer0, boxRep0, viewer1, boxRep1)
        return [bounds[0], bounds[1], bounds[2], bounds[3], 0, 0]

    def _boundsFromViewerToViewer(self, viewer0:Image3DViewer, boxRep0, viewer1:Image3DViewer, boxRep1) -> Sequence[float]:
        bounds0 = boxRep0.GetBounds()
        bounds1 = boxRep1.GetBounds()

        corner0Odd = [bounds0[0], bounds0[2], bounds0[4]]
        corner0Even = [bounds0[1], bounds0[3], bounds0[5]]

        corner1Odd = [bounds1[0], bounds1[2], bounds1[4]]
        corner1Even = [bounds1[1], bounds1[3], bounds1[5]]


        tform0 = vtkCommonMath.vtkMatrix4x4()
        tform0.DeepCopy(viewer0._viewMatrix)

        tform0Inverted = vtkCommonMath.vtkMatrix4x4()
        tform0Inverted.DeepCopy(viewer0._viewMatrix)
        tform0Inverted.Invert()

        tform1 = vtkCommonMath.vtkMatrix4x4()
        tform1.DeepCopy(viewer1._viewMatrix)

        tform1Inverted = vtkCommonMath.vtkMatrix4x4()
        tform1Inverted.DeepCopy(viewer1._viewMatrix)
        tform1Inverted.Invert()

        corner1Odd_2 = tform1.MultiplyPoint((corner1Odd[0], corner1Odd[1], corner1Odd[2], 1))
        corner1Odd_2 = tform0Inverted.MultiplyPoint((corner1Odd_2[0], corner1Odd_2[1], corner1Odd_2[2], 1))

        corner0OddTransformed = tform0.MultiplyPoint((corner0Odd[0], corner0Odd[1], corner1Odd_2[2], 1))

        corner0Odd_2 = tform1Inverted.MultiplyPoint((corner0OddTransformed[0], corner0OddTransformed[1], corner0OddTransformed[2], 1))

        corner1Even_2 = tform1.MultiplyPoint((corner1Even[0], corner1Even[1], corner1Even[2], 1))
        corner1Even_2 = tform0Inverted.MultiplyPoint((corner1Even_2[0], corner1Even_2[1], corner1Even_2[2], 1))

        corner0EvenTransformed = tform0.MultiplyPoint((corner0Even[0], corner0Even[1], corner1Even_2[2], 1))

        corner0Even_2 = tform1Inverted.MultiplyPoint((corner0EvenTransformed[0], corner0EvenTransformed[1], corner0EvenTransformed[2], 1))

        return [corner0Odd_2[0], corner0Even_2[0], corner0Odd_2[1], corner0Even_2[1], corner0Odd_2[2], corner0Even_2[2]]


    def getBoundingBox(self):
        viewer0 = self._viewer0
        viewer1 = self._viewer1

        boxRep0 = self.boxRep0
        boxRep1 = self.boxRep1

        bounds0 = boxRep0.GetBounds()
        bounds1 = boxRep1.GetBounds()

        corner0Odd = [bounds0[0], bounds0[2], bounds0[4]]
        corner0Even = [bounds0[1], bounds0[3], bounds0[5]]

        corner1Odd = [bounds1[0], bounds1[2], bounds1[4]]
        corner1Even = [bounds1[1], bounds1[3], bounds1[5]]

        tform0 = vtkCommonMath.vtkMatrix4x4()
        tform0.DeepCopy(viewer0._viewMatrix)

        tform0Inverted = vtkCommonMath.vtkMatrix4x4()
        tform0Inverted.DeepCopy(viewer0._viewMatrix)
        tform0Inverted.Invert()

        tform1 = vtkCommonMath.vtkMatrix4x4()
        tform1.DeepCopy(viewer1._viewMatrix)

        tform1Inverted = vtkCommonMath.vtkMatrix4x4()
        tform1Inverted.DeepCopy(viewer1._viewMatrix)
        tform1Inverted.Invert()

        corner1Odd_2 = tform1.MultiplyPoint((corner1Odd[0], corner1Odd[1], corner1Odd[2], 1))
        corner1Odd_2 = tform0Inverted.MultiplyPoint((corner1Odd_2[0], corner1Odd_2[1], corner1Odd_2[2], 1))

        corner0OddTransformed = tform0.MultiplyPoint((corner0Odd[0], corner0Odd[1], corner1Odd_2[2], 1))


        corner1Even_2 = tform1.MultiplyPoint((corner1Even[0], corner1Even[1], corner1Even[2], 1))
        corner1Even_2 = tform0Inverted.MultiplyPoint((corner1Even_2[0], corner1Even_2[1], corner1Even_2[2], 1))

        corner0EvenTransformed = tform0.MultiplyPoint((corner0Even[0], corner0Even[1], corner1Even_2[2], 1))

        return [corner0OddTransformed[0], corner0EvenTransformed[0], corner0OddTransformed[1], corner0EvenTransformed[1], corner0OddTransformed[2], corner0EvenTransformed[2]]


from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QLabel


class BlackEmptyPlot(QWidget):
    def __init__(self, parent):
        super().__init__(parent)

        self._layout = QHBoxLayout()
        self.setLayout(self._layout)

        self._layout.addStretch(1)
        self.setContentsMargins(0, 0, 0, 0)

        self.setAttribute(Qt.WA_StyledBackground, True)
        self.setStyleSheet('background-color: black')


from typing import Optional

import vtkmodules.vtkRenderingCore as vtkRenderingCore

from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.lookupTables import fusionLTTo3DLT
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.primaryImage3DLayer_3D import PrimaryImage3DLayer_3D


class SecondaryImage3DLayer_3D(PrimaryImage3DLayer_3D):
    def __init__(self, renderer, renderWindow, iStyle):
        super().__init__(renderer, renderWindow, iStyle)

        self._volumeProperty = vtkRenderingCore.vtkVolumeProperty()
        self._volumeProperty.SetInterpolationTypeToLinear()
        self._volumeProperty.ShadeOn()
        self._volumeProperty.SetAmbient(0.4)
        self._volumeProperty.SetDiffuse(0.6)
        self._volumeProperty.SetSpecular(0.2)


    def _setImage(self, image:Optional[GenericImageForViewer]):
        super()._setImage(image)

        if not (self._image is None):
            self._updateLookupTable(self._image.lookupTable)

    def _updateLookupTable(self, lt):
        volumeColor, volumeScalarOpacity, volumeGradientOpacity = fusionLTTo3DLT(lt)
        self._volumeProperty.SetColor(volumeColor)
        self._volumeProperty.SetScalarOpacity(volumeScalarOpacity)
        self._volumeProperty.SetGradientOpacity(volumeGradientOpacity)

        self._renderWindow.Render()



from PyQt5.QtWidgets import QWidget, QVBoxLayout, QComboBox, QLabel, QPushButton

from opentps.core.data.images import CTImage
from opentps.core.data._patient import Patient
from opentps.core.processing.registration.registrationMorphons import RegistrationMorphons
from opentps.core.processing.registration.registrationDemons import RegistrationDemons
from opentps.core.processing.registration.registrationRigid import RegistrationRigid

class RegistrationPanel(QWidget):
    def __init__(self, viewController):
        QWidget.__init__(self)

        self._patient:Patient = None
        self._viewController = viewController
        self._ctImages = []
        self._selectedFixed = None
        self._selectedMoving = None

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self._fixedLabel = QLabel('Fixed CT:')
        self.layout.addWidget(self._fixedLabel)
        self._fixedComboBox = QComboBox(self)
        self._fixedComboBox.currentIndexChanged.connect(self._handleFixedIndex)
        self.layout.addWidget(self._fixedComboBox)

        self._movingLabel = QLabel('Moving CT:')
        self.layout.addWidget(self._movingLabel)
        self._movingComboBox = QComboBox(self)
        self._movingComboBox.currentIndexChanged.connect(self._handleMovingIndex)
        self.layout.addWidget(self._movingComboBox)

        self._regLabel = QLabel('registration method:')
        self._methods = ['Morphons', 'Demons', 'Rigid']
        self.layout.addWidget(self._regLabel)
        self._regComboBox = QComboBox(self)
        self._regComboBox.currentIndexChanged.connect(self._handleRegIndex)
        self.layout.addWidget(self._regComboBox)
        for reg in self._methods:
            self._regComboBox.addItem(reg)

        self._runButton = QPushButton('Register')
        self._runButton.clicked.connect(self._run)
        self.layout.addWidget(self._runButton)

        self.layout.addStretch()

        self.setCurrentPatient(self._viewController.currentPatient)
        self._viewController.currentPatientChangedSignal.connect(self.setCurrentPatient)

    def _handleFixedIndex(self, *args):
        self._selectedFixed = self._ctImages[self._fixedComboBox.currentIndex()]

    def _handleMovingIndex(self, *args):
        self._selectedMoving = self._ctImages[self._movingComboBox.currentIndex()]

    def _handleRegIndex(self, *args):
        self._selectedReg = self._methods[self._regComboBox.currentIndex()]

    def setCurrentPatient(self, patient:Patient):
        if not (self._patient is None):
            self._patient.imageAddedSignal.disconnect(self._handleImageAddedOrRemoved)
            self._patient.imageRemovedSignal.disconnect(self._handleImageAddedOrRemoved)

        self._patient = patient

        if self._patient is None:
            self._removeAllCTs()
        else:
            self._patient.imageAddedSignal.connect(self._handleImageAddedOrRemoved)
            self._patient.imageRemovedSignal.connect(self._handleImageAddedOrRemoved)

            self._updateCTComboBoxes()

    def _updateCTComboBoxes(self):
        self._removeAllCTs()

        self._ctImages = [ct for ct in self._patient.getPatientDataOfType(CTImage)]

        for ct in self._ctImages:
            self._addFixed(ct)
            self._addMoving(ct)

        try:
            currentIndex = self._ctImages.index(self._selectedFixed)
            self._fixedComboBox.setCurrentIndex(currentIndex)
        except:
            self._fixedComboBox.setCurrentIndex(0)
            if len(self._ctImages):
                self._selectedFixed = self._ctImages[0]

        try:
            currentIndex = self._ctImages.index(self._selectedMoving)
            self._movingComboBox.setCurrentIndex(currentIndex)
        except:
            self._movingComboBox.setCurrentIndex(0)
            if len(self._ctImages):
                self._selectedMoving = self._ctImages[0]

    def _removeAllCTs(self):
        for fixed in self._ctImages:
            self._removeFixed(fixed)
        for moving in self._ctImages:
            self._removeMoving(moving)

    def _addFixed(self, fixed:CTImage):
        self._fixedComboBox.addItem(fixed.name, fixed)
        fixed.nameChangedSignal.connect(self._handleFixedChanged)

    def _addMoving(self, moving:CTImage):
        self._movingComboBox.addItem(moving.name, moving)
        moving.nameChangedSignal.connect(self._handleMovingChanged)

    def _removeFixed(self, fixed:CTImage):
        if fixed==self._selectedFixed:
            self._selectedFixed = None

        fixed.nameChangedSignal.disconnect(self._handleFixedChanged)
        self._fixedComboBox.removeItem(self._fixedComboBox.findData(fixed))

    def _removeMoving(self, moving:CTImage):
        if moving==self._selectedMoving:
            self._selectedMoving = None

        moving.nameChangedSignal.disconnect(self._handleMovingChanged)
        self._movingComboBox.removeItem(self._movingComboBox.findData(moving))

    def _handleImageAddedOrRemoved(self, image):
        self._updateCTComboBoxes()

    def _handleFixedChanged(self, fixed):
        self._updateCTComboBoxes()

    def _handleMovingChanged(self, moving):
        self._updateCTComboBoxes()

    def _run(self):

        if self._selectedReg == 'Morphons':
            reg = RegistrationMorphons(self._selectedFixed, self._selectedMoving)
        elif self._selectedReg == 'Demons':
            reg = RegistrationDemons(self._selectedFixed, self._selectedMoving)
        elif self._selectedReg == 'Rigid':
            reg = RegistrationRigid(self._selectedFixed, self._selectedMoving)
        else:
            print('Not yet implemented')
        reg.compute()
        deformed = reg.deformed
        deformed.patient = self._selectedMoving.patient



from typing import Optional

import vtkmodules.vtkRenderingCore as vtkRenderingCore
from vtkmodules import vtkCommonMath
from vtkmodules.vtkCommonColor import vtkNamedColors
from vtkmodules.vtkIOGeometry import vtkSTLReader
import vtk

from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer


class PrimaryImage2DLayer:
    def __init__(self, renderer, renderWindow, iStyle):
        self.imageChangedSignal = Event(object)

        colors = vtkNamedColors()

        self._image = None
        self._iStyle = iStyle
        self._mainMapper = vtk.vtkImageMapper()
        self._mainActor = vtkRenderingCore.vtkActor2D()
        self._mainActor.SetMapper(self._mainMapper)
        # self._mainMapper = self._mainActor.GetMapper()
        # print('in PrimaryImage2DLayer init', self._mainMapper)
        # self._orientationActor = vtkActor()
        # self._orientationMapper = vtkDataSetMapper()
        # self._orientationWidget = vtkOrientationMarkerWidget()
        self._renderer = renderer
        self._renderWindow = renderWindow
        # self._reslice = vtkImagingCore.vtkImageReslice()
        self._stlReader = vtkSTLReader()
        self._viewMatrix = vtkCommonMath.vtkMatrix4x4()

        # self._mainMapper.SetSliceAtFocalPoint(True)

        # self._orientationActor.SetMapper(self._orientationMapper)
        # self._orientationActor.GetProperty().SetColor(colors.GetColor3d("Silver"))
        # self._orientationMapper.SetInputConnection(self._stlReader.GetOutputPort())
        # self._orientationWidget.SetViewport(0.8, 0.0, 1.0, 0.2)
        # self._orientationWidget.SetCurrentRenderer(self._renderer)
        # self._orientationWidget.SetInteractor(self._renderWindow.GetInteractor())
        # self._orientationWidget.SetOrientationMarker(self._orientationActor)

        # self._reslice.SetOutputDimensionality(2)
        # self._reslice.SetInterpolationModeToNearestNeighbor()

        self._setMainMapperInputConnection()

    def _setMainMapperInputConnection(self):
        if not (self._image is None):
            self._mainMapper.SetInputConnection(self._image.vtkOutputPort)
        # print('in primaryImage2DLayer, setMainMapperInputCOnnection')
        # print(type(self._image))

    def close(self):
        self._disconnectAll()

    @property
    def image(self) -> Optional[Image3DForViewer]:
        """
        Image displayed
        :type:Optional[Image3DForViewer]
        """
        if self._image is None:
            return None

        return self._image

    @image.setter
    def image(self, image:Optional[GenericImageForViewer]):
        self._setImage(image)

    def _setImage(self, image:Optional[GenericImageForViewer]):
        if image == self._image:
            return

        if not (isinstance(image, GenericImageForViewer) or (image is None)):
            return

        self._image = image

        self._disconnectAll()
        self._renderer.RemoveActor(self._mainActor)
        # self._reslice.RemoveAllInputs()

        if not (self._image is None):
            # self._reslice.SetInputConnection(self._image.vtkOutputPort)

            self._renderer.AddActor(self._mainActor)

            self._image.lookupTableName = 'gray'
            # self._setLookupTable()

            self._connectAll()

        self.imageChangedSignal.emit(self._image)

        self._renderWindow.Render()

    def _setLookupTable(self):
        imageProperty = self._mainActor.GetProperty()
        imageProperty.SetLookupTable(self._image.lookupTable)
        imageProperty.UseLookupTableScalarRangeOn()

    def _updateLookupTable(self, lt):
        imageProperty = self._mainActor.GetProperty()
        imageProperty.SetLookupTable(self._image.lookupTable)

        self._renderWindow.Render()

    @property
    def resliceAxes(self):
        """
        Reslice axes
        """
        return self._reslice.GetResliceAxes()

    @resliceAxes.setter
    def resliceAxes(self, resliceAxes):
        self._reslice.SetResliceAxes(resliceAxes)
        self._orientationActor.PokeMatrix(resliceAxes)

    def _connectAll(self):
        self._image.lookupTableChangedSignal.connect(self._updateLookupTable)
        self._image.rangeChangedSignal.connect(self._render)

    def _disconnectAll(self):
        if self._image is None:
            return

        self._image.lookupTableChangedSignal.disconnect(self._updateLookupTable)
        self._image.rangeChangedSignal.connect(self._render)

    def _render(self, *args):
        self._renderWindow.Render()

    def resliceDataFromPhysicalPoint(self, point):
        imageData = self._reslice.GetInput(0)

        ind = [0, 0, 0]
        imageData.TransformPhysicalPointToContinuousIndex(point, ind)
        return imageData.GetScalarComponentAsFloat(int(ind[0]), int(ind[1]), int(ind[2]), 0)



import os

from PyQt5.QtCore import QSize
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QToolBar, QAction, QMenu, QWidgetAction, QPushButton
import opentps.gui.res.icons as IconModule

class DataViewerToolbar(QToolBar):
    def __init__(self, dataViewer):
        super().__init__(dataViewer)

        self._dataViewer = dataViewer

        self.setIconSize(QSize(16, 16))

        # TODO use full path from import opentps_core.res.icons as iconModule
        iconPath = IconModule.__path__[0] + os.path.sep

        self._buttonDVH = QAction(QIcon(iconPath + "dvh.png"), "DVH", self)
        self._buttonDVH.setStatusTip("DVH")
        self._buttonDVH.triggered.connect(self._handleButtonDVH)
        self._buttonDVH.setCheckable(True)

        self._buttonProfile = QAction(QIcon(iconPath + "profile.png"), "Graph", self)
        self._buttonProfile.setStatusTip("Graph")
        self._buttonProfile.triggered.connect(self._handleButtonGraph)
        self._buttonProfile.setCheckable(True)

        self._buttonViewer = QAction(QIcon(iconPath + "x-ray.png"), "Image viewer", self)
        self._buttonViewer.setStatusTip("Image viewer")
        self._buttonViewer.triggered.connect(self._handleButtonViewer)
        self._buttonViewer.setCheckable(True)

        self._buttonViewer_3D = QAction(QIcon(iconPath + "cube.png"), "3D Image viewer", self)
        self._buttonViewer_3D.setStatusTip("3D Image viewer")
        self._buttonViewer_3D.triggered.connect(self._handleButtonViewer_3D)
        self._buttonViewer_3D.setCheckable(True)

        self.addAction(self._buttonViewer)
        self.addAction(self._buttonProfile)
        self.addAction(self._buttonDVH)
        self.addAction(self._buttonViewer_3D)

        self._menuButton = QPushButton("tools", self)
        self._menu = QMenu(self._menuButton)
        self._menuButton.setMenu(self._menu)
        self._menuAction = QWidgetAction(None)
        self._menuAction.setDefaultWidget(self._menuButton)
        self.addAction(self._menuAction)

        self._dataViewer.displayTypeChangedSignal.connect(self._handleDisplayTypeChange)

    @property
    def toolsMenu(self) -> QMenu:
        return self._menu

    def _handleButtonDVH(self, pressed):
        if pressed:
            self._dataViewer.displayType = self._dataViewer.DisplayTypes.DISPLAY_DVH

    def _handleButtonGraph(self, pressed):
        if pressed:
            self._dataViewer.displayType = self._dataViewer.DisplayTypes.DISPLAY_PROFILE

    def _handleButtonViewer(self, pressed):
        if pressed:
            self._dataViewer.displayType = self._dataViewer.DisplayTypes.DISPLAY_IMAGE3D

    def _handleButtonViewer_3D(self, pressed):
        if pressed:
            self._dataViewer.displayType = self._dataViewer.DisplayTypes.DISPLAY_IMAGE3D_3D

    def _handleDisplayTypeChange(self, displayType):
        self._uncheckAllDisplayButton()

        if displayType == self._dataViewer.DisplayTypes.DISPLAY_DVH:
            self._buttonDVH.setChecked(True)
        elif displayType == self._dataViewer.DisplayTypes.DISPLAY_PROFILE:
            self._buttonProfile.setChecked(True)
        elif displayType == self._dataViewer.DisplayTypes.DISPLAY_IMAGE3D:
            self._buttonViewer.setChecked(True)
        elif displayType == self._dataViewer.DisplayTypes.DISPLAY_IMAGE3D_3D:
            self._buttonViewer_3D.setChecked(True)

    def _uncheckAllDisplayButton(self):
        self._buttonDVH.setChecked(False)
        self._buttonProfile.setChecked(False)
        self._buttonViewer.setChecked(False)
        self._buttonViewer_3D.setChecked(False)




import numpy as np
from vtkmodules.vtkIOImage import vtkImageImport

from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer


class Image3DForViewer(GenericImageForViewer):

    def __init__(self, image):
        super().__init__(image)

        if hasattr(self, '_dataImporter'):
            return

        self._dataImporter = vtkImageImport()
        self._selectedPosition = self.data.origin + self.data.gridSizeInWorldUnit/2.
        self._range = (np.min(self.data.imageArray), np.max(self.data.imageArray))

        self.data.dataChangedSignal.connect(self._updateVTKOutputPort)
        self._updateVTKOutputPort()

    def _updateVTKOutputPort(self):
        shape = self.gridSize  ## dataMultiton magic makes all this available here
        imageOrigin = self.origin
        imageSpacing = self.spacing
        imageData = np.swapaxes(self.imageArray, 0, 2)
        num_array = np.array(np.ravel(imageData), dtype=np.float32)

        self._dataImporter.SetNumberOfScalarComponents(1)
        self._dataImporter.SetDataExtent(0, shape[0] - 1, 0, shape[1] - 1, 0, shape[2] - 1)
        self._dataImporter.SetWholeExtent(0, shape[0] - 1, 0, shape[1] - 1, 0, shape[2] - 1)
        self._dataImporter.SetDataSpacing(imageSpacing[0], imageSpacing[1], imageSpacing[2])
        self._dataImporter.SetDataOrigin(imageOrigin[0], imageOrigin[1], imageOrigin[2])
        self._dataImporter.SetDataScalarTypeToFloat()

        data_string = num_array.tobytes()
        self._dataImporter.CopyImportVoidPointer(data_string, len(data_string))

        self._vtkOutputPort = self._dataImporter.GetOutputPort()

    @property
    def vtkOutputPort(self):
        return self._vtkOutputPort



from vtkmodules.vtkIOImage import vtkImageImport

from opentps.core.data.images._image3D import Image3D
from opentps.core import Event
from opentps.gui.viewer.dataForViewer.ROIMaskForViewer import ROIMaskForViewer
from opentps.gui.viewer.dataForViewer.dataMultiton import DataMultiton


class ROIContourForViewer(DataMultiton):
    def __init__(self, roiContour):
        super().__init__(roiContour)

        if hasattr(self, '_dataImporter'):
            return

        self.visibleChangedSignal = Event(bool)

        self._dataImporter = vtkImageImport()
        self._referenceImage = None
        self._visible = False
        self._vtkOutputPort = None
        self._mask = None

        self.colorChangedSignal.connect(self._updateMask)

        self._updateMask()

    def __setattr__(self, key, value):
        super().__setattr__(key, value)

    @property
    def referenceImage(self) -> Image3D:
        return self._referenceImage

    @referenceImage.setter
    def referenceImage(self, image: Image3D):
        if image==self._referenceImage:
            return

        self._referenceImage = image
        self._updateMask()

    @property
    def visible(self) -> bool:
        return self._visible

    @visible.setter
    def visible(self, visible: bool):
        self._visible = visible

        if not(self._mask is None):
            self._mask.visible = visible

        self.visibleChangedSignal.emit(self._visible)

    def asROIMaskForViewer(self) -> ROIMaskForViewer:
        return ROIMaskForViewer(self._mask)

    def _updateMask(self):
        if self._referenceImage is None:
            self._mask = None
            return

        referenceShape = self.referenceImage.gridSize
        referenceOrigin = self.referenceImage.origin
        referenceSpacing = self.referenceImage.spacing

        mask = self.getBinaryMask(origin=referenceOrigin, gridSize=referenceShape, spacing=referenceSpacing)

        if self._mask is None:
            self._mask = mask
        else:
            self._mask.color = mask.color
            self._mask.imageArray = mask.imageArray
            self._mask.origin = mask.origin
            self._mask.spacing = mask.spacing
            self._mask.name = mask.name

    @property
    def vtkOutputPort(self):
        if self._mask is None:
            return None

        return self.asROIMaskForViewer().vtkOutputPort




import vtkmodules.vtkRenderingOpenGL2 #This is necessary to avoid a seg fault
import vtkmodules.vtkRenderingFreeType  #This is necessary to avoid a seg fault
from vtkmodules.vtkRenderingCore import vtkTextActor


class TextLayer:
    def __init__(self, renderer, renderWindow):
        self._primaryText = ['', '', '']
        self._secondaryText = ['', '', '']
        self._renderer = renderer
        self._renderWindow = renderWindow
        self._primaryTextActor = vtkTextActor()
        self._secondaryTextActor = vtkTextActor()
        self._visible = True

        self._primaryTextActor.GetTextProperty().SetFontSize(14)
        self._primaryTextActor.GetTextProperty().SetColor(1, 0.5, 0)
        self._primaryTextActor.GetPositionCoordinate().SetCoordinateSystemToNormalizedViewport()
        self._primaryTextActor.SetPosition(0.02, 0.85)
        self._secondaryTextActor.GetTextProperty().SetJustificationToLeft()
        self._secondaryTextActor.GetTextProperty().SetVerticalJustificationToTop()

        self._secondaryTextActor.GetTextProperty().SetFontSize(14)
        self._secondaryTextActor.GetTextProperty().SetColor(1, 0.5, 0)
        self._secondaryTextActor.GetPositionCoordinate().SetCoordinateSystemToNormalizedViewport()
        self._secondaryTextActor.SetPosition(0.02, 0.02)
        self._secondaryTextActor.GetTextProperty().SetJustificationToLeft()
        self._secondaryTextActor.GetTextProperty().SetVerticalJustificationToBottom()

        self._renderer.AddActor(self._primaryTextActor)
        self._renderer.AddActor(self._secondaryTextActor)

    def close(self):
        pass

    @property
    def primaryText(self):
        return self._primaryText

    def setPrimaryTextLine(self, line, text):
        self._primaryText[line] = text
        self._primaryTextActor.SetInput(self._primaryText[0] + '\n' + self._primaryText[1] + '\n' + self._primaryText[2])
        self._renderWindow.Render()

    @property
    def secondaryText(self):
        return self._secondaryText

    def setSecondaryTextLine(self, line, text):
        self._secondaryText[line] = text
        self._secondaryTextActor.SetInput(
            self._secondaryText[0] + '\n' + self._secondaryText[1] + '\n' + self._secondaryText[2])
        self._renderWindow.Render()

    @property
    def visible(self):
        return self._visible

    @visible.setter
    def visible(self, visible):
        self._primaryTextActor.SetVisibility(visible)
        self._secondaryTextActor.SetVisibility(visible)
        self._visible = visible



from enum import Enum
from typing import Union, Optional

from PyQt5.QtWidgets import QWidget, QVBoxLayout

from opentps.core.data.images import DoseImage
from opentps.core.data.images._image3D import Image3D
from opentps.core.data.images import Image2D
from opentps.core.data.dynamicData._dynamic3DSequence import Dynamic3DSequence
from opentps.core.data.dynamicData._dynamic3DModel import Dynamic3DModel
from opentps.core.data.dynamicData._dynamic2DSequence import Dynamic2DSequence
from opentps.core.data.plan._rtPlan import RTPlan
from opentps.core import Event
from opentps.gui.viewer.dataViewerComponents.dvhViewerActions import DVHViewerActions
from opentps.gui.viewer.dataViewerComponents.image3DViewer import Image3DViewer
from opentps.gui.viewer.dataViewerComponents.dynamicImage3DViewer import DynamicImage3DViewer
from opentps.gui.viewer.dataViewerComponents.image2DViewer import Image2DViewer
from opentps.gui.viewer.dataViewerComponents.dynamicImage2DViewer import DynamicImage2DViewer
from opentps.gui.viewer.dataViewerComponents.image3DViewer_3D import Image3DViewer_3D
from opentps.gui.viewer.dataViewerComponents.imageViewerActions import ImageViewerActions
from opentps.gui.viewer.dataViewerComponents.dataViewerToolbar import DataViewerToolbar
from opentps.gui.viewer.dataViewerComponents.blackEmptyPlot import BlackEmptyPlot
from opentps.gui.viewer.dataViewerComponents.dvhPlot import DVHViewer
from opentps.gui.viewer.dataViewerComponents.profileViewer import ProfileViewer


class DroppedObject:
    """
    This class aims to standardized object dropping
    TODO: we might want to move this to another file
    """
    class DropTypes:
        # DropTypes is not an Enum because Enum does not preserve string type of class attributes. But we want Drop types
        # to be string to be compatible with QMimeData
        IMAGE = 'IMAGE'
        PLAN = 'PLAN'

    def __init__(self, dropType, droppedData):
        self.dropType = dropType
        self.droppedData = droppedData


class DataViewer(QWidget):
    """
    This class displays the type of viewer specified as input and a toolbar to switch between them.
    All viewers are cached for responsiveness.
    Example::
     dataViewer = DataViewer(viewController)
     dataViewer.viewerMode = dataViewer.ViewerModes.STATIC # Static mode (for data which are not time series)
     dataViewer.displayType = dataViewer.ViewerTypes.VIEWER_IMAGE # Show an image viewer
     dataViewe.displayMode = dataViewer.ViewerTypes.VIEWER_DVH # Switch to a DVH viewer
    Currently the DataViewer has its own logic based on events in viewController. Should we have a controller to specifically handle the logical part?
    """
    class DisplayTypes(Enum):
        """
            viewer types
        """
        NONE = None
        DISPLAY_DVH = 'DISPLAY_DVH'
        DISPLAY_PROFILE = 'DISPLAY_PROFILE'
        DISPLAY_IMAGE3D = 'DISPLAY_IMAGE3D'
        DISPLAY_IMAGE2D = 'DISPLAY_IMAGE2D'
        DISPLAY_IMAGE3D_3D = 'DISPLAY_IMAGE3D_3D'

        DEFAULT = DISPLAY_IMAGE3D

    class DisplayModes(Enum):
        """
            viewer modes
        """
        STATIC = 'STATIC'
        DYNAMIC = 'DYNAMIC'

        DEFAULT = STATIC

    class DropModes(Enum):
        AUTO = 'auto'
        PRIMARY = 'primary'
        SECONDARY = 'secondary'

        DEFAULT = AUTO

    def __init__(self, viewController):
        QWidget.__init__(self)

        # It might seems weird to have a signal which is only used within the class but it is if someday we want to move the logical part out of this class.
        self.droppedImageSignal = Event(object)
        self.droppedPlanSignal = Event(object)
        self.displayTypeChangedSignal = Event(object)

        self._viewController = viewController

        self._currentViewer = None
        self._displayMode = self.DisplayModes.DEFAULT
        self._displayType = None

        self._dropMode = self._viewController.dropMode
        self._dropEnabled = False

        self._mainLayout = QVBoxLayout(self)
        self.setLayout(self._mainLayout)
        self._mainLayout.setContentsMargins(0, 0, 0, 0)

        self._toolbar = DataViewerToolbar(self)

        # For responsiveness, we instantiate all possible viewers and hide them == cached viewers:
        self._dvhViewer = DVHViewer(self)
        self._noneViewer = BlackEmptyPlot(self)
        self._staticProfileviewer = ProfileViewer(viewController)
        self._staticImage3DViewer = Image3DViewer(viewController)
        self._staticImage2DViewer = Image2DViewer(viewController)
        self._staticImage3DViewer_3D = Image3DViewer_3D(viewController)

        ## dynamic stuff
        self._dynImage3DViewer = DynamicImage3DViewer(viewController)
        self._dynImage2DViewer = DynamicImage2DViewer(viewController)

        ## hide everything
        self._dvhViewer.hide()
        self._noneViewer.hide()
        self._staticProfileviewer.hide()
        self._staticImage3DViewer.hide()
        self._staticImage2DViewer.hide()
        self._staticImage3DViewer_3D.hide()

        self._dynImage3DViewer.hide()
        self._dynImage2DViewer.hide()

        self._addViewersToLayout()

        self._setDisplayType(self.DisplayTypes.DEFAULT)

        # Logical control of the DataViewer is set here. We might want to move this to dedicated controller class
        self._initializeControl()

    def closeEvent(self, QCloseEvent):
        self.cachedStaticImage3DViewer.close()
        self.cachedStaticImage2DViewer.close()
        self.cachedStaticImage3DViewer_3D.close()
        self.cachedDynamicImage3DViewer.close()
        self.cachedDynamicImage2DViewer.close()
        super().closeEvent(QCloseEvent)

    def _addViewersToLayout(self):
        self._mainLayout.addWidget(self._toolbar)
        self._mainLayout.addWidget(self._dynImage3DViewer)
        self._mainLayout.addWidget(self._dynImage2DViewer)
        self._mainLayout.addWidget(self._staticImage3DViewer)
        self._mainLayout.addWidget(self._staticImage3DViewer_3D)
        self._mainLayout.addWidget(self._staticImage2DViewer)
        self._mainLayout.addWidget(self._staticProfileviewer)
        self._mainLayout.addWidget(self._noneViewer)
        self._mainLayout.addWidget(self._dvhViewer)

    @property
    def cachedDynamicImage3DViewer(self) -> DynamicImage3DViewer:
        """
            The dynamic 3D image viewer currently in cache (read-only)

            :type: Dynamic3DImageViewer
        """
        return self._dynImage3DViewer

    @property
    def cachedDynamicImage2DViewer(self) -> DynamicImage2DViewer:
        """
            The dynamic 2D image viewer currently in cache (read-only)

            :type: Dynamic2DImageViewer
        """
        return self._dynImage2DViewer

    @property
    def cachedStaticDVHViewer(self) -> DVHViewer:
        """
            The DVH viewer currently in cache
        """
        return self._dvhViewer

    @property
    def cachedStaticImage3DViewer(self) -> Image3DViewer:
        """
            The static image 3D viewer currently in cache (read-only)

            :type: Image3DViewer
        """
        return self._staticImage3DViewer

    @property
    def cachedStaticImage3DViewer_3D(self) -> Image3DViewer_3D:
        return self._staticImage3DViewer_3D

    @property
    def cachedStaticImage2DViewer(self) -> Image2DViewer:
        """
            The static image 2D viewer currently in cache (read-only)

            :type: Image2DViewer
        """
        return self._staticImage2DViewer

    @property
    def cachedStaticProfileViewer(self) -> ProfileViewer:
        """
        The profile viewer currently in cache (read-only)

        :type: ProfilePlot
        """
        return self._staticProfileviewer

    @property
    def currentViewer(self) -> Optional[Union[DVHViewer, ProfileViewer, Image3DViewer]]:
        """
        The viewer currently displayed (read-only)viewerTypes

        :type: Optional[Union[DVHViewer, ProfilePlot, ImageViewer]]
        """
        return self._currentViewer

    @property
    def displayType(self):
        """
        The display type of the data viewer tells whether a image viewer, a dvh viewer, ... is displayed

        :type: DataViewer.viewerTypes
        """
        return self._displayType

    @displayType.setter
    def displayType(self, displayType):
        self._setDisplayType(displayType)

    def _setDisplayType(self, displayType):
        if displayType == self._displayType:
            return

        isModeDynamic = self._displayMode == self.DisplayModes.DYNAMIC

        if isModeDynamic:
            self._setDisplayInDynamicMode(displayType)
        else:
            self._setDisplayInStaticMode(displayType)

        self.displayTypeChangedSignal.emit(displayType)

    def setDisplayTypeAndMode(self, args):
        """
        This should be used instead of the displayType and displayMode setters when the type and mode change simultaneously, for example when passing from a 3D dynamic image to a 2D static image or DVH plot

        Parameters
        ----------
        args

        Returns
        -------

        """
        print(NotImplementedError)

        return

    @property
    def displayMode(self):
        """
        The mode of the viewer can be dynamic for dynamic data (time series) or static for static data
        """
        return self._displayMode

    @displayMode.setter
    def displayMode(self, mode):
        if mode == self._displayMode:
            return

        self._disconnectAllViewers()

        # Notify dynamicDisplayController - we have a problem of multiple responsibilities here
        previousModeWasStatic = self.displayMode == self.DisplayModes.STATIC
        if previousModeWasStatic:
            self._viewController.dynamicDisplayController.addDynamicViewer(self.cachedDynamicImage3DViewer)
        else:
            self._viewController.dynamicDisplayController.removeDynamicViewer(self.cachedDynamicImage3DViewer)

        self._displayMode = mode

        # Reset display
        isModeDynamic = self._displayMode == self.DisplayModes.DYNAMIC

        if isModeDynamic:
            self._setDisplayInDynamicMode(self._displayType)
        else:
            self._setDisplayInStaticMode(self._displayType)

    @property
    def dropMode(self):
        return self._dropMode

    @dropMode.setter
    def dropMode(self, mode):
        if mode==self._dropMode:
            return
        self._dropMode = mode

    def _setDisplayInDynamicMode(self, displayType):
        if not (self._currentViewer is None):
            self._currentViewer.hide()

        self._displayType = displayType

        if self._displayType == self.DisplayTypes.DISPLAY_IMAGE3D:
            self._setCurrentViewerToDynamicImage3DViewer()
        elif self._displayType == self.DisplayTypes.DISPLAY_IMAGE2D:
            self._setCurrentViewerToDynamicImage2DViewer()
        elif self._displayType == self.DisplayTypes.DISPLAY_PROFILE:
            self._currentViewer = self._staticProfileviewer
        elif self._displayType == self.DisplayTypes.NONE:
            self._currentViewer = self._noneViewer
        else:
            raise ValueError('Invalid display type: ' + str(self._displayType))

        self._currentViewer.show()

    def _setDisplayInStaticMode(self, displayType):
        if not (self._currentViewer is None):
            self._currentViewer.hide()

        self._displayType = displayType

        if self._displayType == self.DisplayTypes.DISPLAY_DVH:
            self._currentViewer = self._dvhViewer
            self._connectAllViewersInStatic3D()
        elif self._displayType == self.DisplayTypes.NONE:
            self._currentViewer = self._noneViewer
        elif self._displayType == self.DisplayTypes.DISPLAY_PROFILE:
            self._currentViewer = self._staticProfileviewer
            self._connectAllViewersInStatic3D()
        elif self._displayType == self.DisplayTypes.DISPLAY_IMAGE3D:
            self._setCurrentViewerToStaticImage3DViewer()
            self._connectAllViewersInStatic3D()
        elif self._displayType == self.DisplayTypes.DISPLAY_IMAGE2D:
            self._setCurrentViewerToStaticImage2DViewer()
        elif self._displayType == self.DisplayTypes.DISPLAY_IMAGE3D_3D:
            self._setCurrentViewerToStaticImage3DViewer_3D()
            self._connectAllViewersInStatic3D()
        else:
            raise ValueError('Invalid display type: ' + str(self._displayType))

        self._currentViewer.show()

    def _setCurrentViewerToDynamicImage3DViewer(self):
        self._currentViewer = self._dynImage3DViewer
        self.dropEnabled = self._dropEnabled

        self._viewController.crossHairEnabledSignal.connectIfNotAlready(self._dynImage3DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.connectIfNotAlready(self._dynImage3DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.connectIfNotAlready(self._dynImage3DViewer._contourLayer.setNewContour)
        self._viewController.windowLevelEnabledSignal.connectIfNotAlready(self._dynImage3DViewer.setWWLEnabled)

    def _setCurrentViewerToDynamicImage2DViewer(self):
        self._currentViewer = self._dynImage2DViewer
        self.dropEnabled = self._dropEnabled

        self._viewController.crossHairEnabledSignal.connectIfNotAlready(self._dynImage2DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.connectIfNotAlready(self._dynImage2DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.connectIfNotAlready(self._dynImage2DViewer._contourLayer.setNewContour)
        self._viewController.windowLevelEnabledSignal.connectIfNotAlready(self._dynImage2DViewer.setWWLEnabled)

    def _setCurrentViewerToStaticImage3DViewer(self):
        self._currentViewer = self._staticImage3DViewer
        self.dropEnabled = self._dropEnabled

        self._connectStaticImage3DViewer()

    def _connectStaticImage3DViewer(self):
        self._viewController.crossHairEnabledSignal.connectIfNotAlready(self._staticImage3DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.connectIfNotAlready(self._staticImage3DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.connectIfNotAlready(self._staticImage3DViewer._contourLayer.setNewContour)
        self._viewController.windowLevelEnabledSignal.connectIfNotAlready(self._staticImage3DViewer.setWWLEnabled)

    def _connectDVHViewer(self):
        self._viewController.showContourSignal.connectIfNotAlready(self._dvhViewer.appendROI)

    def _setCurrentViewerToStaticImage3DViewer_3D(self):
        self._currentViewer = self._staticImage3DViewer_3D
        self.dropEnabled = self._dropEnabled

        self._connectStaticImage3DViewer_3D()

    def _connectStaticImage3DViewer_3D(self):
        self._viewController.showContourSignal.connectIfNotAlready(self._staticImage3DViewer_3D.setNewContour)

    def _connectAllViewersInStatic3D(self):
        self._connectDVHViewer()
        self._connectStaticImage3DViewer()
        self._connectStaticImage3DViewer_3D()

    def _setCurrentViewerToStaticImage2DViewer(self):
        self._currentViewer = self._staticImage2DViewer
        self.dropEnabled = self._dropEnabled

        self._viewController.crossHairEnabledSignal.connectIfNotAlready(self._staticImage2DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.connectIfNotAlready(self._staticImage2DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.connectIfNotAlready(self._staticImage2DViewer._contourLayer.setNewContour)
        # self._viewController.showContourSignal.connectIfNotAlready(self._dvhViewer.appendROI)
        self._viewController.windowLevelEnabledSignal.connectIfNotAlready(self._staticImage2DViewer.setWWLEnabled)

    def _disconnectAllViewers(self):
        self._viewController.crossHairEnabledSignal.disconnect(self._staticImage3DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.disconnect(self._staticImage3DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.disconnect(self._staticImage3DViewer._contourLayer.setNewContour)
        self._viewController.windowLevelEnabledSignal.disconnect(self._staticImage3DViewer.setWWLEnabled)
        self._viewController.showContourSignal.disconnect(self._dvhViewer.appendROI)

        self._viewController.crossHairEnabledSignal.disconnect(self._dynImage3DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.disconnect(self._dynImage3DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.disconnect(self._dynImage3DViewer._contourLayer.setNewContour)
        self._viewController.windowLevelEnabledSignal.disconnect(self._dynImage3DViewer.setWWLEnabled)

        self._viewController.crossHairEnabledSignal.disconnect(self._staticImage2DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.disconnect(self._staticImage2DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.disconnect(self._staticImage2DViewer._contourLayer.setNewContour)
        self._viewController.windowLevelEnabledSignal.disconnect(self._staticImage2DViewer.setWWLEnabled)

        self._viewController.crossHairEnabledSignal.disconnect(self._dynImage2DViewer.setCrossHairEnabled)
        self._viewController.profileWidgetEnabledSignal.disconnect(self._dynImage2DViewer.setProfileWidgetEnabled)
        self._viewController.showContourSignal.disconnect(self._dynImage2DViewer._contourLayer.setNewContour)
        self._viewController.windowLevelEnabledSignal.disconnect(self._dynImage2DViewer.setWWLEnabled)

        self._viewController.showContourSignal.disconnect(self._staticImage3DViewer_3D.setNewContour)

    @property
    def dropEnabled(self) -> bool:
        """
        Drag and drop enabled

        :type: bool
        """
        return self._dropEnabled

    @dropEnabled.setter
    #TODO: Should not drop be implemented in each specific viewer?
    def dropEnabled(self, enabled: bool):
        self._dropEnabled = enabled

        if enabled:
            self.setAcceptDrops(True)
            self.dragEnterEvent = lambda event: event.accept()
            self.dropEvent = lambda event: self._dropEvent(event)
        else:
            self.setAcceptDrops(False)

    def _dropEvent(self, e):
        """
            Handles a drop in the data viewer
        """
        if e.mimeData().hasText():
            droppedIsImage = e.mimeData().text() == DroppedObject.DropTypes.IMAGE
            droppedIsPlan = e.mimeData().text() == DroppedObject.DropTypes.PLAN

            if droppedIsImage:
                e.accept()
                self.droppedImageSignal.emit(self._viewController.selectedImage)
                return
            elif droppedIsPlan:
                e.accept()
                self.droppedPlanSignal.emit(self._viewController.selectedImage)
                return
        e.ignore()



    ####################################################################################################################
    # This is the logical part of the viewer. Should we migrate this to a dedicated controller?
    def _initializeControl(self):
        self._imageViewerActions = ImageViewerActions(self._staticImage3DViewer)
        self._dvhViewerActions = DVHViewerActions(self._dvhViewer)

        self._imageViewerActions.addToToolbar(self._toolbar)
        self._dvhViewerActions.addToToolbar(self._toolbar)

        self.displayTypeChangedSignal.connectIfNotAlready(self._handleDisplayTypeChange)

        self._viewController.independentViewsEnabledSignal.connectIfNotAlready(self.enableDrop)
        self._viewController.mainImageChangedSignal.connectIfNotAlready(self._setMainImageAnSwitchDisplayModeAndType)
        self._viewController.secondaryImageChangedSignal.connectIfNotAlready(self._setSecondaryImage)
        self._viewController.planChangedSignal.connectIfNotAlready(self._setPlan)
        self._viewController.dropModeSignal.connectIfNotAlready(self._setDropMode)
        self._viewController.droppedDataSignal.connectIfNotAlready(self._setDroppedData)

        self.enableDrop(self._viewController.independentViewsEnabled)

        self._handleDisplayTypeChange(self.displayType) # Initialize with current display type

    def _handleDisplayTypeChange(self, displayType):
        self._imageViewerActions.hide()

        if displayType == self.DisplayTypes.DISPLAY_IMAGE3D:
            self._imageViewerActions.setImageViewer(self._currentViewer)
            self._imageViewerActions.show()
        if displayType == self.DisplayTypes.DISPLAY_IMAGE2D:
            self._imageViewerActions.setImageViewer(self._currentViewer)
            self._imageViewerActions.show()

    def enableDrop(self, enabled):
        self.dropEnabled = enabled

        if enabled:
            # It might seems weird to have a signal connected within the class but it is if someday we want to move the logical part out of this class.
            # See also comment on dropEnabled : Should we implement drop directly in ImageViewer?
            self.droppedImageSignal.connectIfNotAlready(self._setDroppedData)
            self.droppedPlanSignal.connectIfNotAlready(self._setPlan)
        else:
            self.droppedImageSignal.disconnect(self._setDroppedData)
            self.droppedPlanSignal.disconnect(self._setPlan)

    def _setDropMode(self, dropMode):
        self.dropMode = dropMode

    def _setDroppedData(self, data):
        if isinstance(data, RTPlan):
            self._setPlan(data)
            return

        if self._dropMode==self.DropModes.PRIMARY:
            self._setMainImageAnSwitchDisplayModeAndType(data)
        if self._dropMode==self.DropModes.SECONDARY:
            self._setSecondaryImage(data)
        if self._dropMode==self.DropModes.AUTO:
            if isinstance(data, DoseImage):
                self._setSecondaryImage(data)
            else:
                self._setMainImageAnSwitchDisplayModeAndType(data)

    def _setMainImageAnSwitchDisplayModeAndType(self, image):
        """
            Switch display mode according to image type and then display image
        """
        if isinstance(image, Image3D) or isinstance(image, Dynamic3DModel):
            self.displayMode = self.DisplayModes.STATIC
            self.displayType = self.DisplayTypes.DISPLAY_IMAGE3D
        elif isinstance(image, Image2D):
            self.displayMode = self.DisplayModes.STATIC
            self.displayType = self.DisplayTypes.DISPLAY_IMAGE2D
        elif isinstance(image, Dynamic3DSequence):
            self.displayMode = self.DisplayModes.DYNAMIC
            self.displayType = self.DisplayTypes.DISPLAY_IMAGE3D
        elif isinstance(image, Dynamic2DSequence):
            self.displayMode = self.DisplayModes.DYNAMIC
            self.displayType = self.DisplayTypes.DISPLAY_IMAGE2D
        elif image is None:
            pass
        else:
            raise ValueError('Image type not supported')

        self._setMainImage(image)

    def _setMainImage(self, image: Optional[Union[Image3D, Dynamic3DSequence, Image2D, Dynamic2DSequence, Dynamic3DModel]]):
        """
        Set main image to the appropriate cached image viewer.
        Does not affect viewer visibility.
        """
        if isinstance(image, Image3D):
            self.cachedStaticImage3DViewer.primaryImage = image
            self.cachedStaticImage3DViewer_3D.primaryImage = image
            if self.displayMode == self.DisplayModes.DYNAMIC and self.displayType==self.DisplayTypes.DISPLAY_IMAGE3D_3D:
                self.cachedStaticImage3DViewer_3D.update()
        elif isinstance(image, Image2D):
            self.cachedStaticImage2DViewer.primaryImage = image
        elif isinstance(image, Dynamic3DSequence):
            self.cachedDynamicImage3DViewer.primaryImage = image
        elif isinstance(image, Dynamic2DSequence):
            self.cachedDynamicImage2DViewer.primaryImage = image
        elif isinstance(image, Dynamic3DModel):
            self.cachedStaticImage3DViewer.primaryImage = image.midp
        elif image is None:
            if self.displayMode == self.DisplayModes.STATIC and self.displayType==self.DisplayTypes.DISPLAY_IMAGE3D:
                self.cachedStaticImage3DViewer.primaryImage = None
                self.cachedStaticImage3DViewer_3D.primaryImage = image
            elif self.displayMode == self.DisplayModes.STATIC and self.displayType==self.DisplayTypes.DISPLAY_IMAGE2D:
                self.cachedStaticImage2DViewer.primaryImage = None
            elif self.displayMode == self.DisplayModes.DYNAMIC and self.displayType==self.DisplayTypes.DISPLAY_IMAGE3D:
                self.cachedDynamicImage3DViewer.primaryImage = None
            elif self.displayMode == self.DisplayModes.DYNAMIC and self.displayType==self.DisplayTypes.DISPLAY_IMAGE2D:
                self.cachedDynamicImage2DViewer.primaryImage = None

        else:
            raise ValueError('Image type not supported')

        if not image is None and not image.patient is None:
            image.patient.imageRemovedSignal.connectIfNotAlready(self._removeImageFromViewers)

    def _setSecondaryImage(self, image: Optional[Image3D]):
        """
            Display the image (in static mode)
            Does not affect viewer visibility nor viewer type.
        """

        if image is None:
            oldImage = self.cachedStaticImage3DViewer.secondaryImage
            if oldImage is None:
                return
        elif not (image.patient is None):
            image.patient.imageRemovedSignal.connectIfNotAlready(self._removeImageFromViewers)

        self.cachedStaticImage3DViewer.secondaryImage = image
        self.cachedStaticImage3DViewer_3D.secondaryImage = image
        self._setDVHDose(image)

    def _setPlan(self, plan:Optional[RTPlan]):
        self.cachedStaticImage3DViewer.rtPlan = plan
        self.cachedStaticImage3DViewer_3D.rtPlan = plan
        if self.displayMode == self.DisplayModes.DYNAMIC and self.displayType == self.DisplayTypes.DISPLAY_IMAGE3D_3D:
            self.cachedStaticImage3DViewer_3D.update()

    def _setDVHDose(self, image:Optional[DoseImage]):
        self.cachedStaticDVHViewer.dose = image

    def _removeImageFromViewers(self, image: Union[Image3D, ]):
        """
        Remove image from all cached viewers -> The two # lines caused problems because self.cachedStaticImage2DViewer.primaryImage didn't always exist. To be investigated.
        """

        if self.cachedStaticImage3DViewer.primaryImage == image:
            self.cachedStaticImage3DViewer.primaryImage = None
        if self.cachedStaticImage3DViewer.secondaryImage == image:
            self._setSecondaryImage(None)

        if self.cachedDynamicImage3DViewer.primaryImage == image:
            self.cachedStaticImage3DViewer.primaryImage = image

        # if self.cachedStaticImage2DViewer.primaryImage == image:
        if hasattr(self.cachedStaticImage2DViewer, 'primaryImage') and self.cachedStaticImage2DViewer.primaryImage == image:
            self.cachedStaticImage2DViewer.primaryImage = image

        # if self.cachedDynamicImage2DViewer.primaryImage == image:
        if hasattr(self.cachedDynamicImage2DViewer, 'primaryImage') and self.cachedStaticImage2DViewer.primaryImage == image:
            self.cachedDynamicImage2DViewer.primaryImage = None

        if self.cachedStaticDVHViewer.dose == image:
            self._setDVHDose(None)

        if self.cachedStaticImage3DViewer_3D.primaryImage==image:
            self.cachedStaticImage3DViewer_3D.primaryImage = None
        if self.cachedStaticImage3DViewer_3D.secondaryImage==image:
            self.cachedStaticImage3DViewer_3D.secondaryImage = None

        #image.patient.imageRemovedSignal.disconnect(self._removeImageFromViewers)



from PyQt5.QtWidgets import QWidget, QSplitter, QHBoxLayout, QPushButton, QTextEdit, QStatusBar
from PyQt5.QtCore import Qt

from opentps.core import APIInterpreter
from opentps.gui.panels.scriptingPanel.pythonHighlighter import PythonHighlighter


class ScriptingWindow(QWidget):
    def __init__(self):
        QWidget.__init__(self)
        self.setWindowTitle('Script')

        layout = QHBoxLayout()
        self.setLayout(layout)

        horizontalSplitter = QSplitter(Qt.Horizontal)
        verticalSplitter = QSplitter(Qt.Vertical)
        verticalSplitter.addWidget(horizontalSplitter)

        layout.addWidget(verticalSplitter)

        self._runButton = QPushButton('Run')
        self._runButton.clicked.connect(self._runCode)

        self._statusBar = QStatusBar()

        verticalSplitter.addWidget(self._runButton)
        verticalSplitter.addWidget(self._statusBar)

        self._codeTextEdit = QTextEdit()
        PythonHighlighter(self._codeTextEdit)

        self._stdOutput = QTextEdit()
        self._stdOutput.setReadOnly(True)

        horizontalSplitter.addWidget(self._codeTextEdit)
        horizontalSplitter.addWidget(self._stdOutput)


    def _runCode(self):
        try:
            self._statusBar.showMessage("Executing...")

            output = APIInterpreter.run(self._getCode())
            self._stdOutput.setText(output)

            self._statusBar.showMessage("Done.")
        except Exception as err:
            self._statusBar.showMessage(format(err))

    def _getCode(self):
        return self._codeTextEdit.toPlainText()


import functools
import glob
import logging
import os

from PyQt5.QtWidgets import QToolBox, QWidget

from opentps.core import Event
from opentps.gui.panels.doseComparisonPanel import DoseComparisonPanel
from opentps.gui.panels.doseComputationPanel import DoseComputationPanel
from opentps.gui.panels.patientDataPanel.patientDataPanel import PatientDataPanel
from opentps.gui.panels.planDesignPanel.planDesignPanel import PlanDesignPanel
from opentps.gui.panels.planEvaluationPanel import PlanEvaluationPanel
from opentps.gui.panels.planOptimizationPanel.planOptiPanel import PlanOptiPanel
from opentps.gui.panels.roiPanel import ROIPanel
from opentps.gui.panels.scriptingPanel.scriptingPanel import ScriptingPanel
from opentps.gui.panels.registrationPanel import RegistrationPanel

logger = logging.getLogger(__name__)


class MainToolbar(QToolBox):
    class ToolbarItem:
        def __init__(self, panel:QWidget, panelName:str):
            self.visibleEvent = Event(bool)

            self.panel = panel
            self.panelName = panelName
            self.itemNumber = None

            self._visible = True

        @property
        def visible(self) -> bool:
            return self._visible

        @visible.setter
        def visible(self, visible:bool):
            if visible==self._visible:
                return

            self._visible = visible
            self.visibleEvent.emit(self._visible)

    def __init__(self, viewController):
        QToolBox.__init__(self)

        self._viewController = viewController
        self._items = []
        self._maxWidth = 270

        self.setStyleSheet("QToolBox::tab {font: bold; color: #000000; font-size: 16px;}")

        # initialize toolbox panels
        patientDataPanel = PatientDataPanel(self._viewController)
        roiPanel = ROIPanel(self._viewController)
        planDesignPanel = PlanDesignPanel(self._viewController)
        planDesignPanel.setMaximumWidth(self._maxWidth)
        planOptiPanel = PlanOptiPanel(self._viewController)
        planOptiPanel.setMaximumWidth(self._maxWidth)
        dosePanel = DoseComputationPanel(self._viewController)
        dosePanel.setMaximumWidth(self._maxWidth)
        planEvaluationPanel = PlanEvaluationPanel(self._viewController)
        planEvaluationPanel.setEnabled(False)
        doseComparisonPanel = DoseComparisonPanel(self._viewController)
        scriptingPanel = ScriptingPanel()
        #breathingSignalPanel = BreathingSignalPanel(self._viewController)
        #xRayProjPanel = DRRPanel(self._viewController)
        registrationPanel = RegistrationPanel(self._viewController)

        self.addWidget(patientDataPanel, 'Patient data')
        self.addWidget(roiPanel, 'ROI')
        self.addWidget(planDesignPanel, 'Plan design')
        self.addWidget(planOptiPanel, 'Plan optimization')
        self.addWidget(dosePanel, 'Dose computation')
        self.addWidget(planEvaluationPanel, "Plan evaluation")
        self.addWidget(doseComparisonPanel, 'Dose comparison')
        self.addWidget(scriptingPanel, 'Scripting')

    def addWidget(self, widget:QWidget, name:str):
        item = self.ToolbarItem(widget, name)
        self.showItem(item)
        item.visibleEvent.connect(functools.partial(self._handleVisibleEvent, item))

    def _handleVisibleEvent(self, item:ToolbarItem, visible:bool):
        if visible:
            self.showItem(item)
        else:
            self.hideItem(item)

    def showItem(self, item):
        if item in self._items:
            return

        self._items.append(item)
        self.addItem(item.panel, item.panelName)

    def hideItem(self, item):
        if not(item in self._items):
            return

        self.removeItem(self._items.index(item))
        self._items.remove(item)

    @property
    def items(self):
        return [item for item in self._items]



from PyQt5.QtWidgets import QMainWindow, QGroupBox, QHBoxLayout, QWidget, QVBoxLayout, QLabel, QLineEdit
from matplotlib import pyplot as plt
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.widgets import RangeSlider

from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer
from opentps.gui.viewer.dataViewerComponents.patientDataPropertyEditor import PatientDataPropertyEditor


class ImageFusionPropEditor(QMainWindow):
    def __init__(self, image, parent=None):
        super().__init__(parent)

        self.setWindowTitle('Secondary image')
        self.resize(800, 600)

        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        self._layout = QHBoxLayout()
        centralWidget.setLayout(self._layout)

        self._imageInfoGroup = QGroupBox(title='Image info')
        self._layout.addWidget(self._imageInfoGroup)
        vbox = QVBoxLayout()
        self._imageInfoGroup.setLayout(vbox)
        vbox.addWidget(PatientDataPropertyEditor(image, parent=self))

        self._imageProperties = QGroupBox(title='Image properties')
        self._layout.addWidget(self._imageProperties)
        vbox = QVBoxLayout()
        self._imageProperties.setLayout(vbox)


        image = Image3DForViewer(image)

        self._figure = plt.figure()

        axs = plt.axes([0.20, 0.3, 0.60, 0.6])
        n, bins, patches = axs.hist(image.imageArray.flatten(), bins=200)
        axs.set_title('Histogram of pixel intensities')
        axs.set_yscale('log')

        # Create the RangeSlider
        self._slider_ax = plt.axes([0.20, 0.1, 0.60, 0.03])
        self.slider = RangeSlider(self._slider_ax, "Range", bins[0], bins[-1], valinit=image.range, dragging=True)

        self.cm = plt.get_cmap(image.lookupTableName)

        bin_centers = 0.5 * (bins[:-1] + bins[1:])
        col = (bin_centers - image.range[0]) / (image.range[1] - image.range[0])
        for c, p in zip(col, patches):
            plt.setp(p, 'facecolor', self.cm(c))

        self.bin_centers = bin_centers
        self._image = image
        self.patches = patches
        self.slider.on_changed(self._update)

        self._canvas = FigureCanvasQTAgg(self._figure)
        vbox.addWidget(self._canvas)

        self._rangeEditor = RangeEditor(image)
        vbox.addWidget(self._rangeEditor)

        self._image.rangeChangedSignal.connect(self._updateSliderRange)

    def _updateSliderRange(self, range):
        self.slider.set_val(range)

    def _update(self, val):
        self._image.range = val
        col = (self.bin_centers - val[0]) / (val[1] - val[0])
        for c, p in zip(col, self.patches):
            plt.setp(p, 'facecolor', self.cm(c))

class RangeEditor(QWidget):
    def __init__(self, image, parent=None):
        super().__init__(parent)

        self._image = image

        self._mainLayout = QHBoxLayout(self)
        self.setLayout(self._mainLayout)

        self._txt = QLabel(self)
        self._txt.setText('Range: ')

        self._rangeEdit0 = QLineEdit(self)
        self._rangeEdit1 = QLineEdit(self)

        self._mainLayout.addWidget(self._txt)
        self._mainLayout.addWidget(self._rangeEdit0)
        self._mainLayout.addWidget(self._rangeEdit1)

        self._rangeEdit0.setText(str(self._image.range[0]))
        self._rangeEdit1.setText(str(self._image.range[1]))

        self._rangeEdit0.textEdited.connect(self._handleTextEdited)
        self._rangeEdit1.textEdited.connect(self._handleTextEdited)
        self._image.rangeChangedSignal.connect(self.setRangeValue)

    def setRangeValue(self, range):
        if float(self._rangeEdit0.text()) != range[0]:
            self._rangeEdit0.setText(str(range[0]))
        if float(self._rangeEdit1.text()) != range[1]:
            self._rangeEdit1.setText(str(range[1]))

    def _handleTextEdited(self, *args):
        self._image.range = (float(self._rangeEdit0.text()), float(self._rangeEdit1.text()))


from PyQt5 import QtCore
from PyQt5.QtCore import QSize
from PyQt5.QtWidgets import QHBoxLayout, QFrame, QSplitter, QVBoxLayout

from opentps.gui.viewer.doseComparisonDataViewer import DoseComparisonDataViewer
from opentps.gui.viewer.grid import Grid


class GridFourElements(Grid):
    def __init__(self, viewController, parent):
        super().__init__(viewController, parent)

        self._minimumSize = QSize(200, 200)
        self._setEqualSize = False #Use to set equal size before qwidget is effectively shown

        # Horizontal splitting
        self._mainLayout = QHBoxLayout(self)

        self._horizontalSplitter = QSplitter(QtCore.Qt.Horizontal, self)
        self._leftPart = QFrame(self._horizontalSplitter)
        self._leftPart.setFrameShape(QFrame.StyledPanel)
        self._rightPart = QFrame(self._horizontalSplitter)
        self._rightPart.setFrameShape(QFrame.StyledPanel)
        self._horizontalSplitter.addWidget(self._leftPart)
        self._horizontalSplitter.addWidget(self._rightPart)
        self._horizontalSplitter.setStretchFactor(1, 1)

        self._mainLayout.addWidget(self._horizontalSplitter)
        self._mainLayout.setContentsMargins(0, 0, 0, 0)

        # Vertical splitting
        leftPartLayout = QVBoxLayout(self._leftPart)
        leftPartLayout.setContentsMargins(0, 0, 0, 0)
        rightPartLayout = QVBoxLayout(self._rightPart)
        rightPartLayout.setContentsMargins(0, 0, 0, 0)

        self._leftPartSplitter = VerticalSplitter(self)
        self._botLeftGridElementContainer = self._leftPartSplitter.botGridElementContainer
        self._topLeftGridElementContainer = self._leftPartSplitter.topGridElementContainer
        leftPartLayout.addWidget(self._leftPartSplitter)

        self._rightPartSplitter = VerticalSplitter(self)
        self._botRightGridElementContainer = self._rightPartSplitter.botGridElementContainer
        self._topRightGridElementContainer = self._rightPartSplitter.topGridElementContainer
        rightPartLayout.addWidget(self._rightPartSplitter)

        self._botLeftGridElementContainer.setMinimumSize(self._minimumSize)
        self._botRightGridElementContainer.setMinimumSize(self._minimumSize)
        self._topLeftGridElementContainer.setMinimumSize(self._minimumSize)
        self._topRightGridElementContainer.setMinimumSize(self._minimumSize)

        self._botLeftLayout = QVBoxLayout(self._botLeftGridElementContainer)
        self._botRightLayout = QVBoxLayout(self._botRightGridElementContainer)
        self._topLeftLayout = QVBoxLayout(self._topLeftGridElementContainer)
        self._topRightLayout = QVBoxLayout(self._topRightGridElementContainer)

        self._botLeftLayout.setContentsMargins(0, 0, 0, 0)
        self._botRightLayout.setContentsMargins(0, 0, 0, 0)
        self._topLeftLayout.setContentsMargins(0, 0, 0, 0)
        self._topRightLayout.setContentsMargins(0, 0, 0, 0)

        self._initializeViewers()

    def closeEvent(self, QCloseEvent):

        super().closeEvent(QCloseEvent)

    def _initializeViewers(self):
        # Fill grid elements with data viewers
        gridElement = DoseComparisonDataViewer(self._viewController)
        gridElement.cachedStaticImage3DViewer.viewType = gridElement.cachedStaticImage3DViewer.ViewerTypes.AXIAL
        # gridElement.cachedStaticImage2DViewer.viewType = gridElement.cachedStaticImage2DViewer.ViewerTypes.AXIAL
        gridElement.cachedDynamicImage3DViewer.viewType = gridElement.cachedDynamicImage3DViewer.ViewerTypes.AXIAL
        # gridElement.cachedDynamicImage2DViewer.viewType = gridElement.cachedDynamicImage2DViewer.ViewerTypes.AXIAL
        self.appendGridElement(gridElement)
        self._topLeftLayout.addWidget(gridElement)
        gridElement = DoseComparisonDataViewer(self._viewController)
        gridElement.cachedStaticImage3DViewer.viewType = gridElement.cachedStaticImage3DViewer.ViewerTypes.CORONAL
        # gridElement.cachedStaticImage2DViewer.viewType = gridElement.cachedStaticImage2DViewer.ViewerTypes.CORONAL
        gridElement.cachedDynamicImage3DViewer.viewType = gridElement.cachedDynamicImage3DViewer.ViewerTypes.CORONAL
        # gridElement.cachedDynamicImage2DViewer.viewType = gridElement.cachedDynamicImage2DViewer.ViewerTypes.CORONAL
        self.appendGridElement(gridElement)
        self._topRightLayout.addWidget(gridElement)
        gridElement = DoseComparisonDataViewer(self._viewController)
        gridElement.cachedStaticImage3DViewer.viewType = gridElement.cachedStaticImage3DViewer.ViewerTypes.SAGITTAL
        # gridElement.cachedStaticImage2DViewer.viewType = gridElement.cachedStaticImage2DViewer.ViewerTypes.SAGITTAL
        gridElement.cachedDynamicImage3DViewer.viewType = gridElement.cachedDynamicImage3DViewer.ViewerTypes.SAGITTAL
        # gridElement.cachedDynamicImage2DViewer.viewType = gridElement.cachedDynamicImage2DViewer.ViewerTypes.SAGITTAL
        self.appendGridElement(gridElement)
        self._botLeftLayout.addWidget(gridElement)
        gridElement = DoseComparisonDataViewer(self._viewController)
        gridElement.cachedStaticImage3DViewer.viewType = gridElement.cachedStaticImage3DViewer.ViewerTypes.SAGITTAL
        self.appendGridElement(gridElement)
        self._botRightLayout.addWidget(gridElement)


    def resizeEvent(self, event):
        if self._setEqualSize:
            self.setEqualSize()
        self._setEqualSize = False

        self._oldWidth = event.oldSize().width()
        self._newWidth = event.size().width()
        self._leftWidth = self._horizontalSplitter.sizes()[0]

        if self._oldWidth<=0 or self._oldWidth==self.width():
            super().resizeEvent(event)
            return

        oldWidthRatio = self._leftWidth / self._oldWidth

        leftPartWidth = int(oldWidthRatio * self._newWidth)
        rightPartWidth = int(self._newWidth - oldWidthRatio * self._newWidth)

        self._horizontalSplitter.setSizes([leftPartWidth, rightPartWidth])

        super().resizeEvent(event)


    def setEqualSize(self):
        if not self.isVisible():
            self._setEqualSize = True

        leftPartWidth = int(self.width()/2)
        self._horizontalSplitter.setSizes([leftPartWidth, leftPartWidth])

        self._leftPartSplitter.setEqualSize()
        self._rightPartSplitter.setEqualSize()


class VerticalSplitter(QSplitter):
    def __init__(self, parent):
        super().__init__(QtCore.Qt.Vertical, parent)

        self._setEqualSize = False  # Use to set equal size before qwidget is effectively shown

        self.botGridElementContainer = QFrame(self)
        self.topGridElementContainer = QFrame(self)
        self.addWidget(self.topGridElementContainer)
        self.addWidget(self.botGridElementContainer)
        self.setStretchFactor(1, 1)

        self.botGridElementContainer.setFrameShape(QFrame.StyledPanel)
        self.topGridElementContainer.setFrameShape(QFrame.StyledPanel)

    def resizeEvent(self, event):
        if self._setEqualSize:
            self.setEqualSize()
        self._setEqualSize = False

        self._oldHeight = event.oldSize().height()
        self._newHeight = event.size().height()
        self._topHeight = self.sizes()[0]

        if self._oldHeight<=0 or self._oldHeight==self._newHeight:
            super().resizeEvent(event)
            return

        oldHeightRatio = self._topHeight/self._oldHeight

        topHeight = int(oldHeightRatio * self._newHeight)
        botHeight = int(self._newHeight - oldHeightRatio*self._newHeight)

        self.setSizes([topHeight, botHeight])

    def setEqualSize(self):
        if not self.isVisible():
            self._setEqualSize = True

        topHeight = int(self.height()/2)
        self.setSizes([topHeight, topHeight])




import os

from PyQt5.QtCore import QSize, QDir
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QToolBar, QAction, QDialog, QPushButton, QLineEdit, QVBoxLayout, QFileDialog, \
    QStackedWidget, QListView, QComboBox, QWidgetAction

from opentps.core.io import dataLoader
from opentps.core import Event
from opentps.gui.viewer.dataViewer import DataViewer
from opentps.gui.viewer.exportWindow import ExportWindow
from opentps.gui.tools._cropTool import CropWidget
from opentps.gui.tools._resampleTool import ResampleWidget
from opentps.gui.programSettingEditor import ProgramSettingEditor

import opentps.gui.res.icons as IconModule

class ViewerToolbar(QToolBar):

    PLAY_STATUS = 1
    PAUSE_STATUS = 0

    def __init__(self, viewController, parent=None):
        super().__init__(parent)

        self._viewController = viewController
        self._cropWidgets = []
        self._resampleWidgets = []

        self.setIconSize(QSize(16, 16))

        self.iconPath = str(IconModule.__path__[0]) + os.path.sep

        self._buttonSettings = QAction(QIcon(self.iconPath + "settings-5-line.png"), "Settings", self)
        self._buttonSettings.triggered.connect(self._openSettings)

        self._buttonIndependentViews = QAction(QIcon(self.iconPath + "independent_views.png"), "Independent views", self)
        self._buttonIndependentViews.setStatusTip("Independent views")
        self._buttonIndependentViews.triggered.connect(self._handleButtonIndependentViews)
        self._buttonIndependentViews.setCheckable(True)
        self._buttonIndependentViews.setChecked(self._viewController.independentViewsEnabled)

        self._buttonWindowLevel = QAction(QIcon(self.iconPath + "contrast.png"), "Window level", self)
        self._buttonWindowLevel.setStatusTip("Window level")
        self._buttonWindowLevel.triggered.connect(self._handleWindowLevel)
        self._buttonWindowLevel.setCheckable(True)

        self._buttonCrossHair = QAction(QIcon(self.iconPath + "geolocation.png"), "Crosshair", self)
        self._buttonCrossHair.setStatusTip("Crosshair")
        self._buttonCrossHair.triggered.connect(self._handleCrossHair)
        self._buttonCrossHair.setCheckable(True)

        self._buttonCrop = QAction(QIcon(self.iconPath + "crop.png"), "Crop", self)
        self._buttonCrop.setStatusTip("Crop")
        self._buttonCrop.triggered.connect(self._handleCrop)
        self._buttonCrop.setCheckable(False)

        self._buttonResample = QAction(QIcon(self.iconPath + "resample.png"), "Resample", self)
        self._buttonResample.setStatusTip("Resample")
        self._buttonResample.triggered.connect(self._handleResample)
        self._buttonResample.setCheckable(False)

        self._dropModeCombo = QComboBox()
        self._dropModeToStr = {DataViewer.DropModes.AUTO: 'Drop mode: auto',
                               DataViewer.DropModes.PRIMARY: 'Drop as primary image',
                               DataViewer.DropModes.SECONDARY: 'Drop as secondaryImage'}
        self._strToDropMode = {v: k for k, v in self._dropModeToStr.items()}
        self._dropModeCombo.addItems(list(self._dropModeToStr.values()))
        self._dropModeCombo.setCurrentIndex(self._dropModeToIndex(self._viewController.dropMode))
        self._dropModeCombo.currentIndexChanged.connect(self._handleDropModeSelection)
        self._dropModeAction = QWidgetAction(None)
        self._dropModeAction.setDefaultWidget(self._dropModeCombo)

        self.addAction(self._buttonSettings)
        self.addAction(self._buttonIndependentViews)
        self.addAction(self._buttonCrossHair)
        self.addAction(self._buttonWindowLevel)
        self.addAction(self._buttonCrop)
        self.addAction(self._buttonResample)
        self.addAction(self._dropModeAction)

        self.addSeparator()

        ## dynamic options buttons
        self._buttonFaster = QAction(QIcon(self.iconPath + "fast.png"), "Faster", self)
        self._buttonFaster.setStatusTip("Speed up the dynamic viewers evolution")
        self._buttonFaster.triggered.connect(self._handleFaster)

        self._buttonSlower = QAction(QIcon(self.iconPath + "slow.png"), "Slower", self)
        self._buttonSlower.setStatusTip("Slows the dynamic viewers evolution")
        self._buttonSlower.triggered.connect(self._handleSlower)

        self._buttonPlayPause = QAction(QIcon(self.iconPath + "play.png"), "Play/Pause", self)
        self._buttonPlayPause.setStatusTip("Classical Play Pause of course")
        self._buttonPlayPause.triggered.connect(self._handlePlayPause)
        self.playPauseStatus = self.PLAY_STATUS

        self._buttonRefreshRate = QAction('RR', self)
        self._buttonRefreshRate.setStatusTip("Change the refresh rate of the dynamic viewers")
        self._buttonRefreshRate.triggered.connect(self._handleRefreshRate)

        self.fasterSignal = Event()
        self.slowerSignal = Event()
        self.playPauseSignal = Event(bool)
        self.refreshRateChangedSignal = Event(float)
        self.refreshRateValue = 24
        # self.addDynamicButtons()

        self._viewController.independentViewsEnabledSignal.connect(self._handleButtonIndependentViews)
        self._viewController.windowLevelEnabledSignal.connect(self._handleWindowLevel)
        self._viewController.crossHairEnabledSignal.connect(self._handleCrossHair)

    def _dropModeToIndex(self, dropMode):
        return list(self._dropModeToStr.keys()).index(dropMode)

    def _indexToDropMode(self, index):
        return list(self._dropModeToStr.keys())[index]

    def _handleDropModeSelection(self, selectionIndex):
        self._viewController.dropMode = self._indexToDropMode(selectionIndex)

    def _openSettings(self, pressed):
        self._settingEditor = ProgramSettingEditor()
        self._settingEditor.show()

    def _handleButtonIndependentViews(self, pressed):
        # This is useful if controller emit a signal:
        if self._buttonIndependentViews.isChecked() != pressed:
            self._buttonIndependentViews.setChecked(pressed)
            return

        self._viewController.independentViewsEnabled = pressed

    def _handleCrossHair(self, pressed):
        # This is useful if controller emit a signal:
        if self._buttonCrossHair.isChecked() != pressed:
            self._buttonCrossHair.setChecked(pressed)
            return

        self._viewController.crossHairEnabled = pressed

    def _handleCrop(self):
        cw = CropWidget(self._viewController)
        # TODO
        # I don't know why but we must keep the reference to any CropWidget created even if we close it
        # This is an issue because _cropWidgets might become huge
        self._cropWidgets.append(cw)
        cw.show()

    def _handleResample(self):
        rw = ResampleWidget(self._viewController)
        self._resampleWidgets.append(rw)
        rw.show()

    def _handleWindowLevel(self, pressed):
        # This is useful if controller emit a signal:
        if self._buttonWindowLevel.isChecked() != pressed:
            self._buttonWindowLevel.setChecked(pressed)
            return

        self._viewController.windowLevelEnabled = pressed

    def addDynamicButtons(self):

        self.addAction(self._buttonSlower)
        self.addAction(self._buttonPlayPause)
        self._buttonPlayPause.setIcon(QIcon(self.iconPath + "pause.jpg"))
        self.addAction(self._buttonFaster)
        self.addAction(self._buttonRefreshRate)

    def removeDynamicButtons(self):

        self.removeAction(self._buttonSlower)
        self.removeAction(self._buttonPlayPause)
        self.removeAction(self._buttonFaster)
        self.removeAction(self._buttonRefreshRate)

    def _handleFaster(self):
        self.fasterSignal.emit()


    def _handleSlower(self):
        self.slowerSignal.emit()


    def _handlePlayPause(self):
        self.playPauseStatus = not self.playPauseStatus

        if self.playPauseStatus:
            self._buttonPlayPause.setIcon(QIcon(self.iconPath + "pause.jpg"))
        elif not self.playPauseStatus:
            self._buttonPlayPause.setIcon(QIcon(self.iconPath + "play.png"))

        self.playPauseSignal.emit(self.playPauseStatus)


    def _handleRefreshRate(self):

        refreshRateDialog = RefreshRateDialog(self.refreshRateValue)

        if (refreshRateDialog.exec()):
            self.refreshRateValue = float(refreshRateDialog.rRValueLine.text())
            self.refreshRateChangedSignal.emit(self.refreshRateValue)

    # TODO : this is duplicated from patientDataPanel
    def _getOpenFilesAndDirs(self, parent=None, caption='', directory='',
                             filter='', initialFilter='', options=None):
        def updateText():
            # update the contents of the line edit widget with the selected files
            selected = []
            for index in view.selectionModel().selectedRows():
                selected.append('"{}"'.format(index.data()))
            lineEdit.setText(' '.join(selected))

        dialog = QFileDialog(parent, windowTitle=caption)
        dialog.setFileMode(dialog.ExistingFiles)
        if options:
            dialog.setOptions(options)
        dialog.setOption(dialog.DontUseNativeDialog, True)
        if directory:
            dialog.setDirectory(directory)
        if filter:
            dialog.setNameFilter(filter)
            if initialFilter:
                dialog.selectNameFilter(initialFilter)

        # by default, if a directory is opened in file listing mode,
        # QFileDialog.accept() shows the contents of that directory, but we
        # need to be able to "open" directories as we can do with files, so we
        # just override accept() with the default QDialog implementation which
        # will just return exec_()
        dialog.accept = lambda: QDialog.accept(dialog)

        # there are many item views in a non-native dialog, but the ones displaying
        # the actual contents are created inside a QStackedWidget; they are a
        # QTreeView and a QListView, and the tree is only used when the
        # viewMode is set to QFileDialog.Details, which is not this case
        stackedWidget = dialog.findChild(QStackedWidget)
        view = stackedWidget.findChild(QListView)
        view.selectionModel().selectionChanged.connect(updateText)

        lineEdit = dialog.findChild(QLineEdit)
        # clear the line edit contents whenever the current directory changes
        dialog.directoryEntered.connect(lambda: lineEdit.setText(''))

        dialog.exec_()
        return dialog.selectedFiles()


class RefreshRateDialog(QDialog):

    def __init__(self, refreshRateValue):
        QDialog.__init__(self)

        self.RRVal = refreshRateValue

        self.main_layout = QVBoxLayout()
        self.setLayout(self.main_layout)

        # self.slider = QScrollBar(self)   ## to select the refresh rate with a scroll bar, not used for now
        # self.slider.setOrientation(Qt.Horizontal)
        # self.slider.sliderMoved.connect(self.sliderval)
        # self.main_layout.addWidget(self.slider)

        self.rRValueLine = QLineEdit(str(self.RRVal))
        self.main_layout.addWidget(self.rRValueLine)

        okButton = QPushButton("ok", self)
        okButton.clicked.connect(self.accept)
        self.main_layout.addWidget(okButton)


    def sliderval(self):
        # getting current position of the slider
        value = self.slider.sliderPosition()



from vtkmodules.util.vtkAlgorithm import vtkPythonAlgorithm
from vtkmodules.vtkCommonDataModel import vtkImageData

class VtkSimpleImageFilter(vtkPythonAlgorithm):
    def __init__(self):
        super().__init__()
        self.SetNumberOfInputPorts(1)
        self.SetNumberOfOutputPorts(1)

    def RequestData(self, request, inInfoVec, outInfoVec):
        # Get input and output vtkImageData objects
        input_image = vtkImageData.GetData(inInfoVec[0].GetInformationObject(0))
        output_image = vtkImageData.GetData(outInfoVec.GetInformationObject(0))

        # Shallow copy input to output
        output_image.ShallowCopy(input_image)
        return 1




import typing
from enum import Enum

import numpy as np

from opentps.core.data.images import DoseImage
from opentps.core.data.images._image3D import Image3D
from opentps.core.processing.imageProcessing import resampler3D
from opentps.core import Event


class DoseComparisonImageProvider:
    class Metric(Enum):
        ABSOLUTE_DIFFERENCE = "absolute difference"
        DIFFERENCE = "difference"
        GAMMA = "gamma"
        DEFAULT = "difference"

    def __init__(self):
        self.doseComparisonImageChangedSignal = Event(object)

        self._dose1:typing.Optional[DoseImage] = None
        self._dose2:typing.Optional[DoseImage] = None
        self._comparisonMetric = self.Metric.DEFAULT
        self._doseComparisonImage = None

    @property
    def doseComparisonImage(self) -> typing.Optional[Image3D]:
        return self._doseComparisonImage

    @property
    def dose1(self):
        return self._dose1

    @dose1.setter
    def dose1(self, dose:DoseImage):
        self._disconnectAll()
        self._dose1 = dose
        self._updateDoseComparison()
        self._connectAll()

    @property
    def dose2(self):
        return self._dose2

    @dose2.setter
    def dose2(self, dose:DoseImage):
        self._disconnectAll()
        self._dose2 = dose
        self._updateDoseComparison()
        self._connectAll()

    @property
    def comparisonMetric(self):
        return self._comparisonMetric

    @comparisonMetric.setter
    def comparisonMetric(self, m):
        self._comparisonMetric = m
        self._updateDoseComparison()

    def _updateDoseComparison(self):
        if self._dose1 is None:
            dose = None
        elif self._dose2 is None:
            dose = None
        else:
            dose = DoseImage.fromImage3D(self._dose1, patient=None)
            dose2 = resampler3D.resampleImage3DOnImage3D(self._dose2, dose, inPlace=False)
            dose2.patient = None

            if self._comparisonMetric == self.Metric.DIFFERENCE:
                dose.imageArray = dose.imageArray - dose2.imageArray
            elif self._comparisonMetric == self.Metric.ABSOLUTE_DIFFERENCE:
                dose.imageArray = np.abs(dose.imageArray - dose2.imageArray)
            elif self.comparisonMetric == self.Metric.GAMMA:
                raise NotImplementedError

        self._doseComparisonImage = dose
        if not (self._doseComparisonImage is None):
            self._setName()
        self.doseComparisonImageChangedSignal.emit(self._doseComparisonImage)

    def _disconnectAll(self):
        if not (self._dose1 is None):
            self._dose1.nameChangedSignal.disconnect(self._setName)
        if not (self._dose2 is None):
            self._dose2.nameChangedSignal.disconnect(self._setName)

    def _connectAll(self):
        if not (self._dose1 is None):
            self._dose1.nameChangedSignal.connect(self._setName)
        if not (self._dose2 is None):
            self._dose2.nameChangedSignal.connect(self._setName)

    def _setName(self, *args):
        name = ''

        if self._comparisonMetric==self.Metric.DIFFERENCE:
            name = 'Diff. btw. '
        elif self._comparisonMetric==self.Metric.ABSOLUTE_DIFFERENCE:
            name = 'Abs. diff. btw. '
        elif self._comparisonMetric==self.Metric.GAMMA:
            name = 'Gamma ind. btw. '

        name += self._dose1.name
        name += ' and '
        name += self._dose2.name

        self._doseComparisonImage.name = name



import logging
logger = logging.getLogger(__name__)
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QComboBox, QLabel, QLineEdit, QPushButton, QDoubleSpinBox, \
    QHBoxLayout, QCheckBox, QSpinBox
from PyQt5.QtCore import Qt
from opentps.core.data.images import CTImage
from opentps.core.data._patient import Patient
from opentps.core.data._roiContour import ROIContour
from opentps.core.data._rtStruct import RTStruct
from opentps.core.data.plan import RTPlan
from opentps.core.io import mcsquareIO
from opentps.core.io.scannerReader import readScanner
from opentps.core.processing.doseCalculation.doseCalculationConfig import DoseCalculationConfig
from opentps.core.processing.doseCalculation.photons.cccDoseCalculator import CCCDoseCalculator
from opentps.core.processing.doseCalculation.protons.mcsquareDoseCalculator import MCsquareDoseCalculator
from opentps.gui.panels.patientDataWidgets import PatientDataComboBox


class DoseComputationPanel(QWidget):
    def __init__(self, viewController):
        QWidget.__init__(self)

        self._patient:Patient = None
        self._viewController = viewController
        self._ctImages = []
        self._selectedCT = None
        self._rois = []
        self._selectedROI = None
        self._radiationType = "PROTON" # default

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self._radiationLabel = QLabel(self)
        self._radiationLabel.setFixedSize(100, 30)  
        self._radiationLabel.setAlignment(Qt.AlignCenter)  
        self.layout.addWidget(self._radiationLabel)

        self._ctLabel = QLabel('CT:')
        self.layout.addWidget(self._ctLabel)
        self._ctComboBox = PatientDataComboBox(patientDataType=CTImage, patient=self._patient, parent=self)
        self.layout.addWidget(self._ctComboBox)

        self._planLabel = QLabel('Plan:')
        self.layout.addWidget(self._planLabel)
        self._planComboBox = PatientDataComboBox(patientDataType=RTPlan, patient=self._patient, parent=self)
        self.layout.addWidget(self._planComboBox)
        
        self._roiLabel = QLabel('Overwrite outside this ROI:')
        self.layout.addWidget(self._roiLabel)
        self._roiComboBox = QComboBox(self)
        self._roiComboBox.currentIndexChanged.connect(self._handleROIIndex)
        self.layout.addWidget(self._roiComboBox)

        self.layout.addSpacing(15)
        self._doseSpacingLayout = QHBoxLayout()
        self.layout.addLayout(self._doseSpacingLayout)

        self._doseSpacingLabel = QCheckBox('Scoring spacing:')
        self._doseSpacingLabel.toggled.connect(self._setScoringSpacingVisible)
        self._doseSpacingLayout.addWidget(self._doseSpacingLabel)
        self._doseSpacingSpin = QDoubleSpinBox()
        self._doseSpacingSpin.setGroupSeparatorShown(True)
        self._doseSpacingSpin.setRange(0.1, 100.0)
        self._doseSpacingSpin.setSingleStep(1.0)
        self._doseSpacingSpin.setValue(2.0)
        self._doseSpacingSpin.setSuffix(" mm")
        self._doseSpacingLayout.addWidget(self._doseSpacingSpin)
        self._doseSpacingSpin.hide()
        self._doseSpacingLabel.hide()

        self.layout.addSpacing(15)
        self._cropBLBox = QCheckBox('Crop beamlets on ROI')
        self._cropBLBox.setChecked(True)
        self.layout.addWidget(self._cropBLBox)
        self._cropBLBox.hide()

        self.layout.addSpacing(15)
        self._simuProtonLabel = QLabel('<b>Simulation statistics:</b>')
        self.layout.addWidget(self._simuProtonLabel)
        self._numProtons = QSpinBox()
        self._numProtons.setGroupSeparatorShown(True)
        self._numProtons.setRange(0, int(1e9))
        self._numProtons.setSingleStep(int(1e6))
        self._numProtons.setValue(int(1e7))
        self._numProtons.setSuffix(" protons")
        self._simuProtonLabel.hide()
        self._numProtons.hide()

        self._statUncertainty = QDoubleSpinBox()
        self._statUncertainty.setGroupSeparatorShown(True)
        self._statUncertainty.setRange(0.0, 100.0)
        self._statUncertainty.setSingleStep(0.1)
        self._statUncertainty.setValue(2.0)
        self._statUncertainty.setSuffix("% uncertainty")
        self._statUncertainty.hide()
        self.layout.addWidget(self._numProtons)
        self.layout.addWidget(self._statUncertainty)
        self.layout.addSpacing(15)

        self._simuPhotonLabel = QLabel('<b>Simulation parameters:</b>')
        self.layout.addWidget(self._simuPhotonLabel)
        self._batchSize = QSpinBox()
        self._batchSize.setGroupSeparatorShown(True)
        self._batchSize.setRange(0, int(1e9))
        self._batchSize.setSingleStep(10)
        self._batchSize.setValue(30)
        self._batchSize.setSuffix(" (batch size)")
        self.layout.addWidget(self._batchSize)
        self._simuPhotonLabel.hide()
        self._batchSize.hide()


        from opentps.gui.programSettingEditor import MCsquareConfigEditor
        self._mcsquareConfigWidget = MCsquareConfigEditor(self)
        self.layout.addWidget(self._mcsquareConfigWidget)

        self.layout.addSpacing(15)
        self._runButton = QPushButton('Compute dose')
        self._runButton.clicked.connect(self._computeDose)
        self.layout.addWidget(self._runButton)

        self.layout.addStretch()

        self.setCurrentPatient(self._viewController.currentPatient)
        self._viewController.currentPatientChangedSignal.connect(self.setCurrentPatient)
        self._planComboBox.selectedDataEvent.connect(self.updateUIBasedOnRadiationType)
    

    def updateUIBasedOnRadiationType(self,  selectedPlan):
        if selectedPlan is not None:
            if selectedPlan.modality=="RT Plan IOD" and selectedPlan.radiationType.upper()=="PHOTON":
                    self._radiationType = "PHOTON"
                    self._numProtons.hide()
                    self._statUncertainty.hide()
                    self._simuProtonLabel.hide()
                    self._batchSize.show()
                    self._simuPhotonLabel.show()
                    self._mcsquareConfigWidget._txt2.hide()
                    self._mcsquareConfigWidget._bdlField.hide()
                    self._radiationLabel.setText("PHOTON")
                    self._radiationLabel.setStyleSheet("background-color: yellow; color: black;")
            elif selectedPlan.modality=="RT Ion Plan IOD" and selectedPlan.radiationType.upper()=="PROTON":
                self._radiationType = "PROTON"
                self._numProtons.show()
                self._statUncertainty.show()
                self._simuProtonLabel.show()
                self._batchSize.hide()
                self._simuPhotonLabel.hide()
                self._mcsquareConfigWidget._txt2.show()
                self._mcsquareConfigWidget._bdlField.show()
                self._radiationLabel.setText("PROTON")
                self._radiationLabel.setStyleSheet("background-color: red; color: black;")
            else:
                self._radiationLabel.setText("UNKNOWN")
                self._radiationLabel.setStyleSheet("background-color: gray; color: black;")
                logger.warning("Radiation type {} is not supported ".format(selectedPlan.radiationType))

    @property
    def selectedCT(self):
        return self._ctComboBox.selectedData

    @selectedCT.setter
    def selectedCT(self, ct):
        self._ctComboBox.selectedData = ct

    @property
    def selectedPlan(self):
        return self._planComboBox.selectedData

    @selectedPlan.setter
    def selectedPlan(self, plan):
        self._planComboBox.selectedData = plan

    def _handleROIIndex(self, *args):
        self._selectedROI = self._rois[self._roiComboBox.currentIndex()]

    def setCurrentPatient(self, patient:Patient):
        if not (self._patient is None):
            self._patient.rtStructAddedSignal.disconnect(self._handleROIAddedOrRemoved)
            self._patient.rtStructRemovedSignal.disconnect(self._handleROIAddedOrRemoved)

        self._patient = patient

        if self._patient is None:
            self._removeAllCTs()
        else:

            self._patient.rtStructAddedSignal.connect(self._handleROIAddedOrRemoved)
            self._patient.rtStructRemovedSignal.connect(self._handleROIAddedOrRemoved)

            self._planComboBox.setPatient(patient)
            self._ctComboBox.setPatient(patient)

    def _setScoringSpacingVisible(self):
        if self._doseSpacingLabel.isChecked():
            self._doseSpacingSpin.show()
        else:
            self._doseSpacingSpin.hide()

    def _updateROIComboBox(self):
        self._removeAllROIs()

        rtstructs = self._patient.getPatientDataOfType(RTStruct)

        self._rois = []
        for struct in rtstructs:
            for roi in struct:
                self._rois.append(roi)

        for roi in self._rois:
            self._addROI(roi)

        try:
            currentIndex = self._rois.index(self._selectedROI)
            self._roiComboBox.setCurrentIndex(currentIndex)
        except:
            self._roiComboBox.setCurrentIndex(0)
            if len(self._rois):
                self._selectedROI = self._rois[0]

    def _removeAllCTs(self):
        for ct in self._ctImages:
            self._removeCT(ct)

    def _removeAllROIs(self):
        for roi in self._rois:
            self._removeROI(roi)

    def _addROI(self, roi:ROIContour):
        self._roiComboBox.addItem(roi.name, roi)
        roi.nameChangedSignal.connect(self._handleROIChanged)

    def _removeROI(self, roi:ROIContour):
        if roi==self._selectedROI:
            self._selectedROI = None

        roi.nameChangedSignal.disconnect(self._handleROIChanged)
        self._roiComboBox.removeItem(self._roiComboBox.findData(roi))

    def _handleROIAddedOrRemoved(self, roi):
        self._updateROIComboBox()

    def _handleROIChanged(self, roi):
        self._updateROIComboBox()

    def _computeDose(self):
        settings = DoseCalculationConfig()
        calibration = readScanner(settings.scannerFolder)

#        self.selectedPlan.scoringVoxelSpacing = 3 * [self._doseSpacingSpin.value()]
        if self._radiationType.upper()=="PHOTON":
            doseCalculator = CCCDoseCalculator(batchSize= self._batchSize.value())
            doseCalculator.ctCalibration = calibration
            doseImage = doseCalculator.computeDose(self.selectedCT, self.selectedPlan)
            logger.info("Photon dose calculation is done. Check new generated dose image in patient data.")
            doseImage.patient = self.selectedCT.patient
        elif self._radiationType.upper()=="PROTON":
            doseCalculator = MCsquareDoseCalculator()
            beamModel = mcsquareIO.readBDL(settings.bdlFile)
            doseCalculator.beamModel = beamModel
            # self.selectedPlan.scoringVoxelSpacing = self._doseSpacingSpin.value()
            doseCalculator.setScoringParameters(scoringSpacing=self._doseSpacingSpin.value(), adapt_gridSize_to_new_spacing=True)
            doseCalculator.nbPrimaries = self._numProtons.value()
            doseCalculator.statUncertainty = self._statUncertainty.value()
            doseCalculator.ctCalibration = calibration
            doseCalculator.overwriteOutsideROI = self._selectedROI
            doseImage = doseCalculator.computeDose(self.selectedCT, self.selectedPlan)
            logger.info("Proton dose calculation is done. Check new generated dose image in patient data.")
            doseImage.patient = self.selectedCT.patient
        else:
            logger.error("Could not identify plan radiation type")





import logging
import unittest

from PyQt5.QtWidgets import QApplication
from PyQt5.QtWidgets import QMessageBox
from PyQt5.QtGui import QFont

from opentps.core.data import PatientList
from opentps.core.utils.programSettings import ProgramSettings
from opentps.gui.viewController import ViewController


def viewController():
    # instantiate the main opentps_core window
    _viewController = ViewController(patientList)
    _viewController.mainConfig = mainConfig

    return _viewController

logger = logging.getLogger(__name__)

patientList = PatientList()

mainConfig = ProgramSettings()

logger.info("Instantiate opentps gui")
app = QApplication.instance()
if not app:
    app = QApplication([])

def runWithMainWindow(mainWindow):
    # options = parseArgs(sys.argv[1:])
    logger.info("Start opentps gui")

    mainWindow.show()
    #display a message box
    msg = QMessageBox()
    msg.setIcon(QMessageBox.Warning)
    msg.setText(" ------------------------------DISCLAIMER------------------------------ \nOpenTPS is not a medical device, it is purely intended for research purposes and should only be used as such. It is neither FDA or CE approved. The authors of OpenTPS are not responsible for any misuse of the software.")
    msg.setFont(QFont("Arial", 12))
    msg.setWindowTitle("Disclaimer")
    msg.setStandardButtons(QMessageBox.Ok)
    msg.exec_()

    app.exec_()

    mainWindow.close()
    #del mainWindow

def run():
    _viewController = viewController()
    runWithMainWindow(_viewController.mainWindow)

if __name__ == '__main__':
    run()

class MainTestCase(unittest.TestCase):
    def testViewController(self):
        _viewController = viewController()

    def testRun(self):
        print("Testing main window with view controller")
        _viewController = viewController()
        _viewController.mainWindow.show()
        _viewController.mainWindow.close()



from typing import Sequence

from PyQt5 import QtCore
from PyQt5.QtCore import QMimeData, Qt
from PyQt5.QtGui import QStandardItemModel, QDrag, QFont, QStandardItem, QColor
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QComboBox, QTreeView, QAbstractItemView

from opentps.core.data.dynamicData._dynamic2DSequence import Dynamic2DSequence
from opentps.core.data.dynamicData._dynamic3DModel import Dynamic3DModel
from opentps.core.data.dynamicData._dynamic3DSequence import Dynamic3DSequence

from opentps.core.data.images import CTImage
from opentps.core.data.images import DoseImage
from opentps.core.data.images import Image2D
from opentps.core.data.images._image3D import Image3D
from opentps.core.data.images import VectorField3D
from opentps.core.data.plan._rtPlan import RTPlan
from opentps.core.data import PatientData

from opentps.gui.panels.patientDataPanel.patientDataMenu import PatientDataMenu
from opentps.gui.viewer.dataViewer import DroppedObject


class PatientDataSelection(QWidget):
    def __init__(self, viewController, parent=None):
        super().__init__(parent)

        self._viewController = viewController

        self._mainLayout = QVBoxLayout(self)
        self.setLayout(self._mainLayout)

        self.patientBox = PatientComboBox(self._viewController)

        self._patientDataTree = PatientDataTree(self._viewController, self)

        self._mainLayout.addWidget(self.patientBox)
        self._mainLayout.addWidget(self._patientDataTree)

    @property
    def selectedData(self) -> Sequence[PatientData]:
        selected = self._patientDataTree.selectedIndexes()
        selectedData = [self._patientDataTree.model().itemFromIndex(selectedData).data for selectedData in selected]

        return selectedData

class PatientComboBox(QComboBox):
    def __init__(self, viewController):
        QComboBox.__init__(self)

        self._viewController = viewController

        self._viewController.patientAddedSignal.connect(self._addPatient)
        self._viewController.patientRemovedSignal.connect(self._removePatient)
        self._viewController.currentPatientChangedSignal.connect(self._setCurrentPatient)

        self.currentIndexChanged.connect(self._setCurrentPatientInVC)

        self._initialize()

    def closeEvent(self, QCloseEvent):
        self._viewController.patientAddedSignal.disconnect(self._addPatient)
        self._viewController.patientRemovedSignal.disconnect(self._removePatient)
        self._viewController.currentPatientChangedSignal.disconnect(self._setCurrentPatient)

        super().closeEvent(QCloseEvent)

    def _initialize(self):
        for patient in self._viewController.patientList:
            self._addPatient(patient)

    def _addPatient(self, patient):
        name = patient.name
        if name is None:
            name = 'None'

        self.addItem(name, patient)
        if self.count() == 1:
            self._viewController.currentPatient = patient

    def _removePatient(self, patient):
        self.removeItem(self.findData(patient))

    def _setCurrentPatient(self, patient):
        if patient == self.currentData():
            return

        self.setCurrentIndex(self.findData(patient))


    def _setCurrentPatientInVC(self, index):
        self._viewController.currentPatient = self.currentData()


## ------------------------------------------------------------------------------------------
class PatientDataTree(QTreeView):
    def __init__(self, viewController, patientDataPanel):
        QTreeView.__init__(self)

        self._currentPatient = None
        self.patientDataPanel = patientDataPanel
        self._viewController = viewController

        self.setHeaderHidden(True)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.viewport().installEventFilter(self)
        self.customContextMenuRequested.connect(self._handleRightClick)
        self.resizeColumnToContents(0)
        self.doubleClicked.connect(self._handleDoubleClick)
        self.treeModel = QStandardItemModel()
        self.setModel(self.treeModel)
        self.setColumnHidden(1, True)
        self.expandAll()

        self.buildDataTree(self._viewController.currentPatient)
        self._viewController.currentPatientChangedSignal.connect(self.buildDataTree)

        self.setDragDropMode(QAbstractItemView.InternalMove)
        self.setDragEnabled(True)
        self.setAcceptDrops(True)

    def closeEvent(self, QCloseEvent):
        self._disconnectCurrentPatientAndItems()
        self._viewController.currentPatientChangedSignal.disconnect(self.buildDataTree)

        super().closeEvent(QCloseEvent)

    def _appendData(self, data):
        if isinstance(data, Image3D) or isinstance(data, RTPlan) or isinstance(data, Dynamic3DSequence) or isinstance(data, Dynamic2DSequence):
            rootItem = PatientDataItem(data)
            self.rootNode.appendRow(rootItem)

            if isinstance(data, Dynamic3DSequence):
                for image in data.dyn3DImageList:
                    item = PatientDataItem(image)
                    rootItem.appendRow(item)
                self.rootNode.appendRow(rootItem)

            if isinstance(data, Dynamic2DSequence):
                for image in data.dyn2DImageList:
                    item = PatientDataItem(image)
                    rootItem.appendRow(item)
                self.rootNode.appendRow(rootItem)

    def _removeData(self, data):
        items = []

        for row in range(self.model().rowCount()):
            item = self.model().itemFromIndex(self.model().index(row, 0))
            items.append(item)

        for item in items:
            if item.data == data:
                self.rootNode.removeRow(item.row())
                item.disconnectAll() #Do this explicitely to be sure signals are disconnected

    def mouseMoveEvent(self, event):
        drag = QDrag(self)
        mimeData = QMimeData()

        mimeData.setText(DroppedObject.DropTypes.IMAGE)
        drag.setMimeData(mimeData)

        drag.exec_(QtCore.Qt.CopyAction)

    def buildDataTree(self, patient):
        self._disconnectCurrentPatientAndItems()

        self.treeModel.clear()
        self.rootNode = self.treeModel.invisibleRootItem()
        font_b = QFont()
        font_b.setBold(True)

        self._currentPatient = patient

        if self._currentPatient is None:
            return

        self._currentPatient.patientDataAddedSignal.connect(self._appendData)
        self._currentPatient.patientDataRemovedSignal.connect(self._removeData)

        #TODO: Same with other data

        #images
        images = self._currentPatient.images
        for image in images:
            self._appendData(image)

        if len(images) > 0:
            self._viewController.selectedImage = images[0]

        for plan in patient.plans:
            self._appendData(plan)

        # dynamic sequences
        for dynSeq in patient.dynamic3DSequences:
            self._appendData(dynSeq)

        for dynSeq in patient.dynamic2DSequences:
            self._appendData(dynSeq)

        # dynamic models
        for model in self._currentPatient.dynamic3DModels:
            serieRoot = PatientDataItem(model)
            for field in model.deformationList:
                item = PatientDataItem(field)
                serieRoot.appendRow(item)
            self.rootNode.appendRow(serieRoot)

    def _disconnectCurrentPatientAndItems(self):
        # Disconnect signals
        if not (self._currentPatient is None):
            self._currentPatient.patientDataAddedSignal.disconnect(self._appendData)
            self._currentPatient.patientDataRemovedSignal.disconnect(self._removeData)
            # Delete the open ROI by simulate a unselected clic 
            from opentps.gui.panels.roiPanel import ROIItem
            from opentps.gui.viewer.dataForViewer.ROIContourForViewer import ROIContourForViewer
            for i in range(len(self._currentPatient.rtStructs)):
                for j in range(len(self._currentPatient.rtStructs[i]._contours)):
                    contour = ROIContourForViewer(self._currentPatient.rtStructs[i]._contours[j])
                    if ROIContourForViewer(contour).visible == True :
                        checkbox = ROIItem(contour, self._viewController)
                        checkbox.click()
            self._viewController.mainImage = None
            self._viewController.secondaryImage = None
            self._viewController.selectedImage = None

        # Do this explicitely to be sure signals are disconnected
        for row in range(self.model().rowCount()):
            item = self.model().itemFromIndex(self.model().index(row, 0))
            if isinstance(item, PatientDataItem):
                item.disconnectAll()

    def dragEnterEvent(self, event):
        selection = self.selectionModel().selectedIndexes()[0]
        self._viewController.selectedImage = self.model().itemFromIndex(selection).data

    def _handleDoubleClick(self, selection):
        selectedData = self.model().itemFromIndex(selection).data

        if isinstance(selectedData, CTImage) or isinstance(selectedData, Dynamic3DSequence) or isinstance(selectedData, Dynamic2DSequence) or isinstance(selectedData, Image2D):
            self._viewController.mainImage = selectedData
        if isinstance(selectedData, RTPlan):
            self._viewController.plan = selectedData
        elif isinstance(selectedData, Dynamic3DModel):
            self._viewController.mainImage = selectedData.midp
        elif isinstance(selectedData, DoseImage):
            self._viewController.secondaryImage = selectedData

    def _handleRightClick(self, pos):
        pos = self.mapToGlobal(pos)

        selected = self.selectedIndexes()
        selectedData = [self.model().itemFromIndex(selectedData).data for selectedData in selected]

        dataMenu = PatientDataMenu(self._viewController, self)
        dataMenu.selectedData = selectedData
        dataMenu.asContextMenu().popup(pos)



## ------------------------------------------------------------------------------------------
class PatientDataItem(QStandardItem):
    def __init__(self, data, txt="", dataType="", color=QColor(75, 75, 75)):
        QStandardItem.__init__(self)


        # print('in patientDataSelection.py, PatientDataItem init', type(data), data.getTypeAsString())
        self.data = data
        # print(data.__dict__)
        self.data.nameChangedSignal.connect(self.setName)

        self.setEditable(False)

        defaultColor = color
        dynSeqColor = QColor(109, 119, 125)
        planColor = QColor(56, 130, 176)
        doseColor = QColor(166, 63, 104)
        modelColor = QColor(194, 130, 35)
        vectorFieldColor = QColor(201, 111, 50)

        if isinstance(data, CTImage):
            self.setName('CT' + ' - ' + self.data.name)
            self.setForeground(defaultColor)
        elif isinstance(data, Dynamic3DSequence):
            self.setName('3D Seq' + ' - ' + self.data.name)
            self.setForeground(dynSeqColor)
        elif isinstance(data, Dynamic2DSequence):
            self.setName('2D Seq' + ' - ' + self.data.name)
            self.setForeground(dynSeqColor)
        elif isinstance(data, Image2D):
            self.setName('2D Img' + ' - ' + self.data.name)
            self.setForeground(defaultColor)
        elif isinstance(data, RTPlan):
            self.setName('plan' + ' - ' + self.data.name)
            self.setForeground(planColor)
        elif isinstance(data, Dynamic3DModel):
            self.setName('DynMod' + ' - ' + self.data.name)
            self.setForeground(modelColor)
        elif isinstance(data, VectorField3D):
            self.setName('Vec Field' + ' - ' + self.data.name)
            self.setForeground(vectorFieldColor)
        elif isinstance(data, DoseImage):
            self.setName('Dose' + ' - ' + self.data.name)
            self.setForeground(doseColor)
        else:
            self.setName(data.getTypeAsString() + ' - ' + self.data.name)
            self.setForeground(defaultColor)

        # self.setText(txt)
        # self.setWhatsThis(type)

    def disconnectAll(self):
        self.data.nameChangedSignal.disconnect(self.setName)

    # No effect: it seems that C destructor of QStandardItem does not trigger __del__
    def __del__(self):
        self.disconnectAll()

    def setName(self, name):
        self.setText(name)



from opentps.gui.viewer.dataViewerComponents.image3DViewer import Image3DViewer
from opentps.gui.viewer.dataForViewer.dyn3DSeqForViewer import Dyn3DSeqForViewer


class DynamicImage3DViewer(Image3DViewer):
    def __init__(self, viewController):
        super().__init__(viewController)

        self._viewController = viewController

        self.dynPrimaryImgSeq = None
        self.dynPrimaryImgSeqForViewer = None

        self.dynSecondaryImgSeq = None
        self.dynSecondaryImgSeqForViewer = None

        self.dynContourImgSeq = None
        self.dynContourImgSeqForViewer = None

        self.curPrimaryImgIdx = 0
        self.curSecondaryImgIdx = 0
        self.curContourImgIdx = 0

        self.loopStepNumber = 0

    @property
    def primaryImage(self):
        if self._primaryImageLayer.image is None:
            return None
        return self.dynPrimaryImgSeqForViewer

    @primaryImage.setter
    def primaryImage(self, dyn3DImgSeq):
        if dyn3DImgSeq is None:
            self.dynPrimaryImgSeq = None
            self.dynPrimaryImgSeqForViewer = None
            super().image = None
        elif dyn3DImgSeq != self.dynPrimaryImgSeq:
            self.dynPrimaryImgSeq = dyn3DImgSeq
            self.dynPrimaryImgSeqForViewer = Dyn3DSeqForViewer(self.dynPrimaryImgSeq)
            super()._setPrimaryImageForViewer(self.dynPrimaryImgSeqForViewer)

    def nextImage(self, index):
        self.curPrimaryImgIdx = index
        self.dynPrimaryImgSeqForViewer.currentIndexIn3DSeq = index
        self._renderWindow.Render()

    @property
    def secondaryImage(self):
        return None

    def resetDynamicParameters(self):
        self.curPrimaryImgIdx = 0
        self.curSecondaryImgIdx = 0
        self.curContourImgIdx = 0

        self.loopStepNumber = 0


from typing import Optional

import vtkmodules.vtkRenderingCore as vtkRenderingCore
from vtkmodules.vtkRenderingVolumeOpenGL2 import vtkSmartVolumeMapper

from opentps.core import Event
from opentps.gui.viewer.dataForViewer.genericImageForViewer import GenericImageForViewer
from opentps.gui.viewer.dataForViewer.image3DForViewer import Image3DForViewer
from opentps.gui.viewer.dataViewerComponents.imageViewerComponents.lookupTables import ct3DLT


class PrimaryImage3DLayer_3D:
    def __init__(self, renderer, renderWindow, iStyle):
        self.imageChangedSignal = Event(object)

        self._image = None
        self._imageToBeSet = None
        self._iStyle = iStyle
        self._mainActor = vtkRenderingCore.vtkVolume()
        self._mainMapper = vtkSmartVolumeMapper()
        self._renderer = renderer
        self._renderWindow = renderWindow

        self._mainActor.SetMapper(self._mainMapper)

        self._volumeColor, self._volumeScalarOpacity, self._volumeGradientOpacity = ct3DLT()

        self._volumeProperty = vtkRenderingCore.vtkVolumeProperty()
        self._volumeProperty.SetColor(self._volumeColor)
        self._volumeProperty.SetScalarOpacity(self._volumeScalarOpacity)
        self._volumeProperty.SetGradientOpacity(self._volumeGradientOpacity)
        self._volumeProperty.SetInterpolationTypeToLinear()
        self._volumeProperty.ShadeOn()
        self._volumeProperty.SetAmbient(0.4)
        self._volumeProperty.SetDiffuse(0.6)
        self._volumeProperty.SetSpecular(0.2)

    def close(self):
        self._disconnectAll()
        self._renderer.RemoveActor(self._mainActor)
        self._mainMapper.RemoveAllInputs()

    def update(self):
        self._setImage(self._imageToBeSet)

    @property
    def image(self) -> Optional[Image3DForViewer]:
        return self._imageToBeSet

    @image.setter
    def image(self, image:Optional[GenericImageForViewer]):
        if image == self._imageToBeSet:
            return

        self._imageToBeSet = image

    def _setImage(self, image:Optional[GenericImageForViewer]):
        if self._image == self._imageToBeSet:
            return

        self._image = image

        self._disconnectAll()
        self._renderer.RemoveActor(self._mainActor)
        self._mainMapper.RemoveAllInputs()

        if not (self._image is None):
            self._mainMapper.SetInputConnection(self._image.vtkOutputPort)

            self._renderer.AddActor(self._mainActor)

            self._mainActor.SetProperty(self._volumeProperty)

            self._connectAll()

            self._renderer.ResetCamera()

        self.imageChangedSignal.emit(self._image)

        self._renderWindow.Render()

    def _connectAll(self):
        self._image.dataChangedSignal.connect(self._render)
        self._image.lookupTableChangedSignal.connect(self._updateLookupTable)

    def _disconnectAll(self):
        if self._image is None:
            return

        self._image.dataChangedSignal.disconnect(self._render)
        self._image.lookupTableChangedSignal.disconnect(self._updateLookupTable)

    def _render(self, *args):
        self._renderWindow.Render()

    def _updateLookupTable(self, lt):
        pass




class DataMultiton(object):
    _allViewerDatas = []

    def __new__(cls, data):
        if data is None:
            return None

        # if there is already a DataMultiton for this data instance
        for dataMultiton in DataMultiton._allViewerDatas:
            if dataMultiton.data == data:
                dataMultiton.subClass(cls)
                return dataMultiton

        # else
        dataMultiton = super().__new__(cls)
        DataMultiton._allViewerDatas.append(dataMultiton)
        return dataMultiton

    def subClass(self, cls):
        self.__class__ = cls  # Subclass the existing DataMultiton

    def __init__(self, data):
        if data is None:
            return

        if hasattr(self, 'data'):
            return

        super().__init__()
        self.data = data

    def __getattribute__(self, item):
        try:
            return object.__getattribute__(self, item)
        except:
            # we cannot return data from data
            if item=='data':
                raise AttributeError("data not initialized in DataMultiton")
            return self.data.__getattribute__(item)




__docformat__ = "restructuredtext en"
__all__ = ['ResampleWidget']

from typing import Sequence

import numpy as np
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QMainWindow, QVBoxLayout, QPushButton, QFrame, QLabel, QDoubleSpinBox

from opentps.core.processing.imageProcessing import resampler3D
from opentps.gui.panels.patientDataPanel.patientDataSelection import PatientDataSelection


class ResampleWidget(QMainWindow):
    def __init__(self, viewController, parent=None):
        super().__init__(parent)

        self._viewController = viewController

        self.setWindowTitle('Crop tool')

        centralWidget = QWidget()
        self.setCentralWidget(centralWidget)
        self._mainLayout = QHBoxLayout()
        centralWidget.setLayout(self._mainLayout)

        self._resampleOptions = ResampleOptions(self._viewController)

        self._cropDataButton = QPushButton('Resample all selected data')
        self._cropDataButton.clicked.connect(self._resampleData)

        self._dataSelection = PatientDataSelection((self._viewController))

        self._menuFrame = QFrame(self)
        self._menuFrame.setMaximumWidth(400)
        self._menuLayout = QVBoxLayout(self._menuFrame)
        self._menuFrame.setLayout(self._menuLayout)

        self._mainLayout.addWidget(self._menuFrame)
        self._menuLayout.addWidget(self._dataSelection)
        self._menuLayout.addWidget(self._resampleOptions)
        self._menuLayout.addWidget(self._cropDataButton)

    def _resampleData(self):
        newSpacing = np.array(self._resampleOptions.newSpacing)

        for data in self._dataSelection.selectedData:
            resampler3D.resampleImage3D(data, newSpacing, inPlace=True)


class ResampleOptions(QWidget):
    def __init__(self, viewController):
        super().__init__()

        self._viewController = viewController

        self._mainLayout = QVBoxLayout()
        self.setLayout(self._mainLayout)

        self._layerLabel = QLabel('New spacing:')
        self._mainLayout.addWidget(self._layerLabel)
        self._spacing1Spin = QDoubleSpinBox()

        self._spacingFrame = QFrame(self)
        self._mainLayout.addWidget(self._spacingFrame)
        self._spacingFrame.setFixedWidth(200)
        self._spacingLayout = QHBoxLayout(self._spacingFrame)
        self._spacingFrame.setLayout(self._spacingLayout)

        self._spacing1Spin.setGroupSeparatorShown(True)
        self._spacing1Spin.setRange(0.1, 100.0)
        self._spacing1Spin.setSingleStep(1.0)
        self._spacing1Spin.setValue(2.0)
        self._spacing1Spin.setSuffix(" mm")
        self._spacingLayout.addWidget(self._spacing1Spin)

        self._spacing2Spin = QDoubleSpinBox()
        self._spacing2Spin.setGroupSeparatorShown(True)
        self._spacing2Spin.setRange(0.1, 100.0)
        self._spacing2Spin.setSingleStep(1.0)
        self._spacing2Spin.setValue(2.0)
        self._spacing2Spin.setSuffix(" mm")
        self._spacingLayout.addWidget(self._spacing2Spin)

        self._spacing3Spin = QDoubleSpinBox()
        self._spacing3Spin.setGroupSeparatorShown(True)
        self._spacing3Spin.setRange(0.1, 100.0)
        self._spacing3Spin.setSingleStep(1.0)
        self._spacing3Spin.setValue(2.0)
        self._spacing3Spin.setSuffix(" mm")
        self._spacingLayout.addWidget(self._spacing3Spin)

    @property
    def newSpacing(self) -> Sequence[float]:
        return (self._spacing1Spin.value(), self._spacing2Spin.value(), self._spacing3Spin.value())




from PyQt5.QtCore import QDir
from PyQt5.QtWidgets import QWidget, QVBoxLayout, QPushButton, QFileDialog, QDialog, \
    QStackedWidget, QListView, QLineEdit, QHBoxLayout, QCheckBox

import opentps.core.io.dataLoader as dataLoader
from opentps.core.io.serializedObjectIO import saveDataStructure
from opentps.core import Event
from opentps.gui.viewer.exportWindow import ExportWindow
from opentps.gui.panels.patientDataPanel.patientDataSelection import PatientDataSelection


class PatientDataPanel(QWidget):
    def __init__(self, viewController):
        QWidget.__init__(self)

        # Events
        self.patientAddedSignal = Event(object)
        self.patientRemovedSignal = Event(object)

        self._viewController = viewController

        self._viewController.patientAddedSignal.connect(self.patientAddedSignal.emit)

        self._viewController.patientRemovedSignal.connect(self.patientRemovedSignal.emit)

        self.layout = QVBoxLayout()
        self.setLayout(self.layout)

        self._patientDataSelection = PatientDataSelection(self._viewController)
        self.layout.addWidget(self._patientDataSelection)

        self.buttonLayout = QHBoxLayout()
        loadDataButton = QPushButton('Load data')
        loadDataButton.clicked.connect(self.loadData)
        self.buttonLayout.addWidget(loadDataButton)
        saveDataButton = QPushButton('Save data')
        saveDataButton.clicked.connect(self.saveData)
        self.buttonLayout.addWidget(saveDataButton)
        self.layout.addLayout(self.buttonLayout)

        self.dataPath = QDir.currentPath() # maybe not the ideal default data directory

    def loadData(self):
        filesOrFoldersList = _getOpenFilesAndDirs(caption="Open patient data files or folders", directory=QDir.currentPath())
        if len(filesOrFoldersList) < 1:
            return

        splitPath = filesOrFoldersList[0].split('/')
        withoutLastElementPath = ''
        for element in splitPath[:-1]:
            withoutLastElementPath += element + '/'
        self.dataPath = withoutLastElementPath

        dataLoader.loadData(self._viewController._patientList, filesOrFoldersList)

    def saveData(self):
        self._exportWindow = ExportWindow(self._viewController)
        self._exportWindow.show()

## ------------------------------------------------------------------------------------------
class SaveData_dialog(QFileDialog):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # self.defaultName = "Patient"

    def getSaveFileName(self, parent=None,
                        caption="Select folder and file name to save data tree",
                        dir=".",
                        filter='',
                        initialFilter='',
                        defaultName="Patient",
                        options=None):

        self.setWindowTitle(caption)
        self.setDirectory(dir)
        self.selectFile(defaultName)
        self.setAcceptMode(QFileDialog.AcceptSave)  # bouton "Save"
        self.setOption(QFileDialog.DontUseNativeDialog, True)

        layout = self.layout()
        # checkBoxLayout = QHBoxLayout
        self.compressBox = QCheckBox("Compress data", self)
        self.compressBox.setToolTip('This will compress the data before saving them, it takes longer to save the data this way')
        layout.addWidget(self.compressBox, 4, 0)
        self.splitPatientsBox = QCheckBox("Split Patients", self)
        self.splitPatientsBox.setToolTip('This will split patients into multiple files if multiple patients data are loaded')
        layout.addWidget(self.splitPatientsBox, 4, 1)
        self.setLayout(layout)

        if self.exec_():
            return self.selectedFiles()[0], self.compressBox.isChecked(), self.splitPatientsBox.isChecked()
        else:
            return "", ""

def _getOpenFilesAndDirs(parent=None, caption='', directory='',
                          filter='', initialFilter='', options=None):
    def updateText():
      # update the contents of the line edit widget with the selected files
      selected = []
      for index in view.selectionModel().selectedRows():
        selected.append('"{}"'.format(index.data()))
      lineEdit.setText(' '.join(selected))

    dialog = QFileDialog(parent, windowTitle=caption)
    dialog.setFileMode(dialog.ExistingFiles)
    if options:
      dialog.setOptions(options)
    dialog.setOption(dialog.DontUseNativeDialog, True)
    if directory:
      dialog.setDirectory(directory)
    if filter:
      dialog.setNameFilter(filter)
      if initialFilter:
        dialog.selectNameFilter(initialFilter)

    # by default, if a directory is opened in file listing mode,
    # QFileDialog.accept() shows the contents of that directory, but we
    # need to be able to "open" directories as we can do with files, so we
    # just override accept() with the default QDialog implementation which
    # will just return exec_()
    dialog.accept = lambda: QDialog.accept(dialog)

    # there are many item views in a non-native dialog, but the ones displaying
    # the actual contents are created inside a QStackedWidget; they are a
    # QTreeView and a QListView, and the tree is only used when the
    # viewMode is set to QFileDialog.Details, which is not this case
    stackedWidget = dialog.findChild(QStackedWidget)
    view = stackedWidget.findChild(QListView)
    view.selectionModel().selectionChanged.connect(updateText)

    lineEdit = dialog.findChild(QLineEdit)
    # clear the line edit contents whenever the current directory changes
    dialog.directoryEntered.connect(lambda: lineEdit.setText(''))

    dialog.exec_()
    return dialog.selectedFiles()